"""This contains common utilities to enable components of the prefect imaging flowws.
The majority of the items here are the task decorated functions. Effort should be made
to avoid putting in too many items that are not going to be directly used by prefect
imaging flows.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Collection, Literal, ParamSpec, TypeVar

import numpy as np
import pandas as pd
from prefect import Task, task, unmapped
from prefect.artifacts import create_table_artifact

from flint.calibrate.aocalibrate import (
    ApplySolutions,
    CalibrateCommand,
    create_apply_solutions_cmd,
    select_aosolution_for_ms,
)
from flint.coadd.linmos import LinmosOptions, LinmosResult, linmos_images
from flint.convol import (
    BeamShape,
    convolve_cubes,
    convolve_images,
    get_common_beam,
    get_cube_common_beam,
)
from flint.flagging import flag_ms_aoflagger
from flint.imager.wsclean import (
    ImageSet,
    WSCleanOptions,
    WSCleanResult,
    combine_images_to_cube,
    image_set_from_result,
    merge_image_sets,
    merge_image_sets_from_results,
    split_and_get_image_set,
    wsclean_imager,
)
from flint.logging import logger
from flint.masking import (
    MaskingOptions,
    create_snr_mask_from_fits,
    extract_beam_mask_from_mosaic,
)
from flint.ms import (
    MS,
    copy_and_preprocess_casda_askap_ms,
    preprocess_askap_ms,
    rename_column_in_ms,
    split_by_field,
)
from flint.naming import (
    FITSMaskNames,
    get_beam_resolution_str,
    get_fits_cube_from_paths,
)
from flint.options import FieldOptions, SubtractFieldOptions
from flint.peel.potato import potato_peel
from flint.prefect.common.utils import upload_image_as_artifact
from flint.selfcal.casa import gaincal_applycal_ms
from flint.source_finding.aegean import AegeanOutputs, run_bane_and_aegean
from flint.summary import FieldSummary
from flint.utils import (
    flatten_items,
    zip_folder,
)
from flint.validation import (
    ValidationTables,
    XMatchTables,
    create_validation_plot,
    create_validation_tables,
)

P = ParamSpec("P")
R = TypeVar("R")

# These are simple task wrapped functions and require no other modification
task_copy_and_preprocess_casda_askap_ms: Task[P, R] = task(
    copy_and_preprocess_casda_askap_ms
)
task_preprocess_askap_ms: Task[P, R] = task(preprocess_askap_ms)
task_split_by_field: Task[P, R] = task(split_by_field)
task_select_solution_for_ms: Task[P, R] = task(select_aosolution_for_ms)
task_create_apply_solutions_cmd: Task[P, R] = task(create_apply_solutions_cmd)
task_rename_column_in_ms: Task[P, R] = task(rename_column_in_ms)
task_convolve_images = task(convolve_images)
task_split_and_get_image_set = task(split_and_get_image_set)
task_image_set_from_result = task(image_set_from_result)
task_combine_images_to_cube = task(combine_images_to_cube)
task_merge_image_sets = task(merge_image_sets)
task_merge_image_sets_from_results = task(merge_image_sets_from_results)

# Tasks below are extracting componented from earlier stages, or are
# otherwise doing something important

FlagMS = TypeVar("FlagMS", MS, ApplySolutions)


@task
def task_get_channel_images_from_paths(paths: list[Path]) -> list[Path]:
    return [path for path in paths if "MFS" not in path.name]


@task
def task_potato_peel(
    ms: MS,
    potato_container: Path,
    update_potato_config_options: dict[str, Any] | None = None,
    update_potato_peel_options: dict[str, Any] | None = None,
    update_wsclean_options: dict[str, Any] | None = None,
) -> MS:
    logger.info(f"Attempting to peel {ms.path}")

    if update_wsclean_options is None:
        update_wsclean_options = {}

    wsclean_options = WSCleanOptions(**update_wsclean_options)

    initial_data_column = ms.column

    ms = potato_peel(
        ms=ms,
        potato_container=potato_container,
        update_potato_config_options=update_potato_config_options,
        update_potato_peel_options=update_potato_peel_options,
        image_options=wsclean_options,
    )

    post_data_column = ms.column

    logger.info(f"Initial potato data column: {initial_data_column}")
    logger.info(f"Post potato data column: {post_data_column}")

    if post_data_column != initial_data_column:
        logger.critical(f"{ms.path} data column has changed!")

    return ms


@task
def task_flag_ms_aoflagger(ms: FlagMS, container: Path) -> FlagMS:
    # Pirate believes the type ignore below is due to the decorated function and type alias
    extracted_ms = MS.cast(ms=ms)  # type: ignore

    extracted_ms = flag_ms_aoflagger(ms=extracted_ms, container=container)

    return ms


@task
def task_bandpass_create_apply_solutions_cmd(
    ms: MS, calibrate_cmd: CalibrateCommand, container: Path
) -> ApplySolutions:
    """Apply a ao-calibrate solutions file to the subject measurement set.

    Args:
        ms (MS): The measurement set that will have the solutions file applied
        calibrate_cmd (CalibrateCommand): A resulting ao-calibrate command and meta-data item
        container (Path): The container that can apply the ao-calibrate style solutions file to the measurement set

    Returns:
        ApplySolutions: The resulting apply solutions command and meta-data
    """
    return create_apply_solutions_cmd(
        ms=ms, solutions_file=calibrate_cmd.solution_path, container=container
    )


@task
def task_extract_solution_path(calibrate_cmd: CalibrateCommand) -> Path:
    """Extract the solution path from a calibrate command. This is often required when
    interacting with a ``PrefectFuture`` wrapped ``CalibrateCommand`` result.

    Args:
        calibrate_cmd (CalibrateCommand): The subject calibrate command the solution file will be extracted from

    Returns:
        Path: Path to the solution file
    """
    # TODO: What is actually using this task? Is it better to just pass through a
    # CalibrateCommand?
    # TODO: Use the get attribute task enabled function
    return calibrate_cmd.solution_path


@task
def task_run_bane_and_aegean(
    image: WSCleanResult | LinmosResult,
    aegean_container: Path,
    update_bane_options: dict[str, Any] | None = None,
    update_aegean_options: dict[str, Any] | None = None,
) -> AegeanOutputs:
    """Run BANE and Aegean against a FITS image.

    Notes:
        It has been noted that BANE can sometimes get caught in a interpolation error which haults execution.
        The ``timelimit_seconds`` will attempt to detect long runnings BANE processes and raise an error. The
        retry functionality of prefect should then restart the task. Since this task is pure (e.g. no last
        dataproducts, modification to data etc) simply restarting should be fine.

    Args:
        image (Union[WSCleanResult, LinmosResult]): The image that will be searched
        aegean_container (Path): Path to a singularity container containing BANE and aegean
        update_bane_options (dict[str, Any] | None, optional): Over-ride any default options of BANEOptions. If None defaults are used. Defaults to None.
        update_aegean_options (dict[str, Any] | None, optional): Over-ride any default options of AegeanOptions. If None defaults are used. Defaults to None.

    Raises:
        ValueError: Raised when ``image`` is not a supported type

    Returns:
        AegeanOutputs: Output BANE and aegean products, including the RMS and BKG images
    """
    if isinstance(image, WSCleanResult):
        assert image.image_set is not None, "Image set attribute unset. "
        image_paths = image.image_set.image

        logger.info(f"Have extracted image: {image_paths}")

        # For the moment, will only source find on an MFS image
        image_paths = [image for image in image_paths if ".MFS." in str(image)]
        assert len(image_paths) == 1, (
            "More than one image found after filter_str for MFS only images. "
        )
        # Get out the only path in the list.
        image_path = image_paths[0]
    elif isinstance(image, LinmosResult):
        logger.info("Will be running aegean on a linmos image")

        image_path = image.image_fits
        assert image_path.exists(), f"Image path {image_path} does not exist"
    else:
        raise ValueError(f"Unexpected type, have received {type(image)} for {image=}. ")

    aegean_outputs = run_bane_and_aegean(
        image=image_path,
        aegean_container=aegean_container,
        update_bane_options=update_bane_options,
        update_aegean_options=update_aegean_options,
    )

    return aegean_outputs


@task
def task_zip_ms(in_item: WSCleanResult) -> Path:
    """Zip a measurement set

    Args:
        in_item (WSCleanResult): The inpute item with a ``.ms`` attribute of type ``MS``.

    Returns:
        Path: Output path of the zipped measurement set
    """
    # TODO: This typing needs to be expanded
    ms = in_item.ms

    zipped_ms = zip_folder(in_path=ms.path)

    return zipped_ms


@task
def task_gaincal_applycal_ms(
    ms: MS | WSCleanResult,
    selfcal_round: int,
    casa_container: Path,
    update_gain_cal_options: dict[str, Any] | None = None,
    archive_input_ms: bool = False,
    skip_selfcal: bool = False,
    rename_ms: bool = False,
    archive_cal_table: bool = False,
) -> MS:
    """Perform self-calibration using CASA gaincal and applycal.

    Args:
        ms (Union[MS, WSCleanResult]): A resulting wsclean output. This is used purely to extract the ``.ms`` attribute.
        selfcal_round (int): Counter indication which self-calibration round is being performed. A name is included based on this.
        casa_container (Path): A path to a singularity container with CASA tooling.
        update_gain_cal_options (Optional[Dict[str, Any]], optional): Options used to overwrite the default ``gaincal`` options. Defaults to None.
        archive_input_ms (bool, optional): If True the input measurement set is zipped. Defaults to False.
        skip_selfcal (bool, optional): Should this self-cal be skipped. If `True`, the a new MS is created but not calibrated the appropriate new name and returned.
        rename_ms (bool, optional): It `True` simply rename a MS and adjust columns appropriately (potentially deleting them) instead of copying the complete MS. If `True` `archive_input_ms` is ignored. Defaults to False.
        archive_cal_table (bool, optional): Archive the output calibration table in a tarball. Defaults to False.

    Raises:
        ValueError: Raised when a ``.ms`` attribute can not be obtained

    Returns:
        MS: Self-calibrated measurement set
    """
    # TODO: Need to do a better type system to include the .ms
    # TODO: This needs to be expanded to handle multiple MS
    ms = ms if isinstance(ms, MS) else ms.ms  # type: ignore

    if not isinstance(ms, MS):
        raise ValueError(
            f"Unsupported {type(ms)=} {ms=}. Likely multiple MS instances? This is not yet supported. "
        )

    return gaincal_applycal_ms(
        ms=ms,
        round=selfcal_round,
        casa_container=casa_container,
        update_gain_cal_options=update_gain_cal_options,
        archive_input_ms=archive_input_ms,
        skip_selfcal=skip_selfcal,
        rename_ms=rename_ms,
        archive_cal_table=archive_cal_table,
    )


@task
def task_wsclean_imager(
    in_ms: ApplySolutions | MS,
    wsclean_container: Path,
    update_wsclean_options: dict[str, Any] | None = None,
    fits_mask: FITSMaskNames | None = None,
    channel_range: tuple[int, int] | None = None,
    make_cube_from_subbands: bool = True,
) -> WSCleanResult:
    """Run the wsclean imager against an input measurement set

    Args:
        in_ms (Union[ApplySolutions, MS]): The measurement set that will be imaged
        wsclean_container (Path): Path to a singularity container with wsclean packages
        update_wsclean_options (Optional[Dict[str, Any]], optional): Options to update from the default wsclean options. Defaults to None.
        fits_mask (Optional[FITSMaskNames], optional): A path to a clean guard mask. Defaults to None.
        channel_range (Optional[Tuple[int,int]], optional): Add to the wsclean options the specific channel range to be imaged. Defaults to None.

    Returns:
        WSCleanResult: A resulting wsclean command and resulting meta-data
    """
    from flint.exceptions import CleanDivergenceError

    ms = in_ms if isinstance(in_ms, MS) else in_ms.ms

    update_wsclean_options = (
        {} if update_wsclean_options is None else update_wsclean_options
    )

    if fits_mask:
        update_wsclean_options["fits_mask"] = fits_mask.mask_fits

    if channel_range:
        update_wsclean_options["channel_range"] = channel_range

    logger.info(f"wsclean inager {ms=}")
    try:
        return wsclean_imager(
            ms=ms,
            wsclean_container=wsclean_container,
            update_wsclean_options=update_wsclean_options,
            make_cube_from_subbands=make_cube_from_subbands,
        )
    except CleanDivergenceError:
        # NOTE: If the cleaning failed retry with some larger images
        # and slower cleaning. Perhaps this should be moved closer
        # to the wscleaning functionality
        size = (
            update_wsclean_options["size"] + 1024
            if "size" in update_wsclean_options
            else 8196
        )
        mgain = (
            max(0, update_wsclean_options["mgain"] - 0.1)
            if "mgain" in update_wsclean_options
            else 0.6
        )
        convergence_wsclean_options = dict(size=size, mgain=mgain)
        # dicts are mutable. Don't want to change for everything. Unclear to me
        # how prefect would behave here.
        update_wsclean_options = update_wsclean_options.copy()
        update_wsclean_options.update(**convergence_wsclean_options)
        logger.warning(
            f"Clean divergence dertected. Rerunning. Updated options {convergence_wsclean_options=}"
        )

        return wsclean_imager(
            ms=ms,
            wsclean_container=wsclean_container,
            update_wsclean_options=update_wsclean_options,
            make_cube_from_subbands=make_cube_from_subbands,
        )


def get_common_beam_from_images(
    image_paths: list[Path],
    cutoff: float = 25,
    filter_str: str | None = None,
    fixed_beam_shape: list[float] | None = None,
) -> BeamShape:
    """Compute a common beam size that all input images will be convoled to.

    Args:
        image_paths (list[Path]): Input images whose restoring beam properties will be considered
        cutoff (float, optional): Major axis larger than this valur, in arcseconds, will be ignored. Defaults to 25.
        filter_str (Optional[str], optional): Only include images when considering beam shape if this string is in the file path. Defaults to None.
        fixed_beam_shape (Optional[List[float]], optional): Specify the final beamsize of linmos field images in (arcsec, arcsec, deg). If None it is deduced from images. Defaults to None;

    Returns:
        BeamShape: The final convolving beam size to be used
    """
    # TODO: This function could have a wrapper around it that checks to see if
    # fixed_beam_shape is present, and simply return, avoiding using this functions
    # .submit method. Ahhh.
    if fixed_beam_shape:
        beam_shape = BeamShape(
            bmaj_arcsec=fixed_beam_shape[0],
            bmin_arcsec=fixed_beam_shape[1],
            bpa_deg=fixed_beam_shape[2],
        )
        logger.info(f"Using fixed {beam_shape=}")
        return beam_shape

    if filter_str:
        image_paths = [image for image in image_paths if filter_str in str(image)]

    logger.info(f"Considering {len(image_paths)} images. ")

    beam_shape = get_common_beam(image_paths=image_paths, cutoff=cutoff)
    if np.isnan(beam_shape.bmaj_arcsec):
        logger.critical("Failed to get beam resolution for:")
        logger.critical(f"{image_paths=}")
        logger.critical(f"{cutoff=}")
        logger.critical(f"{filter_str=}")

    return beam_shape


task_get_common_beam_from_images: Task[P, R] = task(get_common_beam_from_images)


def get_common_beam_from_image_set(
    image_set: ImageSet,
    cutoff: float = 25,
    filter_str: str | None = None,
    fixed_beam_shape: list[float] | None = None,
) -> BeamShape:
    """Compute a common beam size that all input images will be convoled to.

    Args:
        wsclean_results (Collection[WSCleanResult]): Input images whose restoring beam properties will be considered
        cutoff (float, optional): Major axis larger than this valur, in arcseconds, will be ignored. Defaults to 25.
        filter_str (Optional[str], optional): Only include images when considering beam shape if this string is in the file path. Defaults to None.
        fixed_beam_shape (Optional[List[float]], optional): Specify the final beamsize of linmos field images in (arcsec, arcsec, deg). If None it is deduced from images. Defaults to None;

    Returns:
        BeamShape: The final convolving beam size to be used
    """
    return get_common_beam_from_images(
        image_paths=image_set.image,
        cutoff=cutoff,
        filter_str=filter_str,
        fixed_beam_shape=fixed_beam_shape,
    )


task_get_common_beam_from_image_set: Task[P, R] = task(get_common_beam_from_image_set)


def get_common_beam_from_image_sets(
    image_sets: list[ImageSet],
    cutoff: float = 25,
    filter_str: str | None = None,
    fixed_beam_shape: list[float] | None = None,
) -> BeamShape:
    """Compute a common beam size that all input images will be convoled to.

    Args:
        wsclean_results (Collection[WSCleanResult]): Input images whose restoring beam properties will be considered
        cutoff (float, optional): Major axis larger than this valur, in arcseconds, will be ignored. Defaults to 25.
        filter_str (Optional[str], optional): Only include images when considering beam shape if this string is in the file path. Defaults to None.
        fixed_beam_shape (Optional[List[float]], optional): Specify the final beamsize of linmos field images in (arcsec, arcsec, deg). If None it is deduced from images. Defaults to None;

    Returns:
        BeamShape: The final convolving beam size to be used
    """
    images_to_consider: list[Path] = []
    for image_set in image_sets:
        images_to_consider.extend(image_set.image)
    return get_common_beam_from_images(
        image_paths=images_to_consider,
        cutoff=cutoff,
        filter_str=filter_str,
        fixed_beam_shape=fixed_beam_shape,
    )


task_get_common_beam_from_image_sets: Task[P, R] = task(get_common_beam_from_image_sets)


def get_common_beam_from_results(
    wsclean_results: list[WSCleanResult],
    cutoff: float = 25,
    filter_str: str | None = None,
    fixed_beam_shape: list[float] | None = None,
) -> BeamShape:
    """Compute a common beam size that all input images will be convoled to.

    Args:
        wsclean_results (Collection[WSCleanResult]): Input images whose restoring beam properties will be considered
        cutoff (float, optional): Major axis larger than this valur, in arcseconds, will be ignored. Defaults to 25.
        filter_str (Optional[str], optional): Only include images when considering beam shape if this string is in the file path. Defaults to None.
        fixed_beam_shape (Optional[List[float]], optional): Specify the final beamsize of linmos field images in (arcsec, arcsec, deg). If None it is deduced from images. Defaults to None;

    Returns:
        BeamShape: The final convolving beam size to be used
    """
    images_to_consider: list[Path] = []

    # TODO: This should support other image types
    for wsclean_result in wsclean_results:
        if wsclean_result.image_set is None:
            logger.warning(
                f"No image_set for {wsclean_result.ms} found. Has imager finished?"
            )
            continue
        images_to_consider.extend(wsclean_result.image_set.image)

    return get_common_beam_from_images(
        image_paths=images_to_consider,
        cutoff=cutoff,
        filter_str=filter_str,
        fixed_beam_shape=fixed_beam_shape,
    )


task_get_common_beam_from_results: Task[P, R] = task(get_common_beam_from_results)


@task
def task_get_cube_common_beam(
    wsclean_results: Collection[WSCleanResult],
    cutoff: float = 25,
) -> list[BeamShape]:
    """Compute a common beam size  for input cubes.

    Args:
        wsclean_results (Collection[WSCleanResult]): Input images whose restoring beam properties will be considered
        cutoff (float, optional): Major axis larger than this valur, in arcseconds, will be ignored. Defaults to 25.

    Returns:
        List[BeamShape]: The final convolving beam size to be used per channel in cubes
    """

    images_to_consider: list[Path] = []

    # TODO: This should support other image types
    for wsclean_result in wsclean_results:
        if wsclean_result.image_set is None:
            logger.warning(
                f"No image_set for {wsclean_result.ms} found. Has imager finished?"
            )
            continue
        images_to_consider.extend(wsclean_result.image_set.image)

    images_to_consider = get_fits_cube_from_paths(paths=images_to_consider)

    logger.info(
        f"Considering {len(images_to_consider)} images across {len(wsclean_results)} outputs. "
    )

    beam_shapes = get_cube_common_beam(cube_paths=images_to_consider, cutoff=cutoff)

    return beam_shapes


@task
def task_convolve_cube(
    wsclean_result: WSCleanResult,
    beam_shapes: list[BeamShape],
    cutoff: float = 60,
    mode: Literal["image"] = "image",
    convol_suffix_str: str = "conv",
) -> Collection[Path]:
    """Convolve images to a specified resolution

    Args:
        wsclean_result (WSCleanResult): Collection of output images from wsclean that will be convolved
        beam_shapes (BeamShape): The shape images will be convolved to
        cutoff (float, optional): Maximum major beam axis an image is allowed to have before it will not be convolved. Defaults to 60.
        convol_suffix_str (str, optional): The suffix added to the convolved images. Defaults to 'conv'.

    Returns:
        Collection[Path]: Path to the output images that have been convolved.
    """
    assert wsclean_result.image_set is not None, (
        f"{wsclean_result.ms} has no attached image_set."
    )

    supported_modes = ("image",)
    logger.info(f"Extracting {mode}")
    if mode == "image":
        image_paths = list(wsclean_result.image_set.image)
    else:
        raise ValueError(f"{mode=} is not supported. Known modes are {supported_modes}")

    logger.info(f"Extracting cubes from image_set {mode=}")
    image_paths = get_fits_cube_from_paths(paths=image_paths)

    # It is possible depending on how aggressively cleaning image products are deleted that these
    # some cleaning products (e.g. residuals) do not exist. There are a number of ways one could consider
    # handling this. The pirate in me feels like less is more, so an error will be enough. Keeping
    # things simple and avoiding the problem is probably the better way of dealing with this
    # situation. In time this would mean that we inspect and handle conflicting pipeline options.
    assert image_paths is not None, (
        f"{image_paths=} for {mode=} and {wsclean_result.image_set=}"
    )

    logger.info(f"Will convolve {image_paths}")

    return convolve_cubes(
        cube_paths=image_paths,
        beam_shapes=beam_shapes,
        cutoff=cutoff,
        convol_suffix=convol_suffix_str,
    )


def convolve_image_set(
    image_set: ImageSet,
    beam_shape: BeamShape,
    cutoff: float = 60,
    mode: str = "image",
    filter_str: str | None = None,
    convol_suffix_str: str = "conv",
    remove_original_images: bool = False,
) -> list[Path]:
    """Convolve images to a specified resolution

    Args:
        wsclean_result (WSCleanResult): Collection of output images from wsclean that will be convolved
        beam_shape (BeamShape): The shape images will be convolved to
        cutoff (float, optional): Maximum major beam axis an image is allowed to have before it will not be convolved. Defaults to 60.
        filter_str (Optional[str], optional): This string must be contained in the image path for it to be convolved. Defaults to None.
        convol_suffix_str (str, optional): The suffix added to the convolved images. Defaults to 'conv'.
        remove_original_images (bool, optional): If True remove the original image after they have been convolved. Defaults to False.

    Returns:
        Collection[Path]: Path to the output images that have been convolved.
    """
    supported_modes = ("image", "residual")
    logger.info(f"Extracting {mode}")
    if mode == "image":
        image_paths = list(image_set.image)
    elif mode == "residual":
        assert image_set.residual is not None, (
            f"{image_set.residual=}, which should not happen"
        )
        image_paths = list(image_set.residual)
    else:
        raise ValueError(f"{mode=} is not supported. Known modes are {supported_modes}")

    if filter_str is not None:
        logger.info(f"Filtering images paths with {filter_str=}")
        image_paths = [
            image_path for image_path in image_paths if filter_str in str(image_path)
        ]

    # It is possible depending on how aggressively cleaning image products are deleted that these
    # some cleaning products (e.g. residuals) do not exist. There are a number of ways one could consider
    # handling this. The pirate in me feels like less is more, so an error will be enough. Keeping
    # things simple and avoiding the problem is probably the better way of dealing with this
    # situation. In time this would mean that we inspect and handle conflicting pipeline options.
    assert image_paths is not None, f"{image_paths=} for {mode=} and {image_set=}"

    logger.info(f"Will convolve {image_paths}")

    # experience has shown that astropy units do not always work correctly
    # in a multiprocessing / dask environment. The unit registry does not
    # seem to serialise correctly, and we can get weird arcsecond is not
    # compatible with arcsecond type errors. Import here in an attempt
    # to minimise
    import astropy.units as u
    from astropy.io import fits
    from radio_beam import Beam

    # Print the beams out here for logging
    for image_path in image_paths:
        image_beam = Beam.from_fits_header(fits.getheader(str(image_path)))
        logger.info(
            f"{image_path.name!s}: {image_beam.major.to(u.arcsecond)} {image_beam.minor.to(u.arcsecond)}  {image_beam.pa}"
        )

    convolved_images = convolve_images(
        image_paths=image_paths,
        beam_shape=beam_shape,
        cutoff=cutoff,
        convol_suffix=convol_suffix_str,
    )

    if remove_original_images:
        logger.info(f"Removing {len(image_paths)} input images")
        _ = [image_path.unlink() for image_path in image_paths]  # type: ignore

    return convolved_images


task_convolve_image_set: Task[P, R] = task(convolve_image_set)


@task
def task_convolve_image(
    wsclean_result: WSCleanResult,
    beam_shape: BeamShape,
    cutoff: float = 60,
    mode: str = "image",
    filter_str: str | None = None,
    convol_suffix_str: str = "conv",
    remove_original_images: bool = False,
) -> Collection[Path]:
    """Convolve images to a specified resolution

    Args:
        wsclean_result (WSCleanResult): Collection of output images from wsclean that will be convolved
        beam_shape (BeamShape): The shape images will be convolved to
        cutoff (float, optional): Maximum major beam axis an image is allowed to have before it will not be convolved. Defaults to 60.
        filter_str (Optional[str], optional): This string must be contained in the image path for it to be convolved. Defaults to None.
        convol_suffix_str (str, optional): The suffix added to the convolved images. Defaults to 'conv'.
        remove_original_images (bool, optional): If True remove the original image after they have been convolved. Defaults to False.

    Returns:
        Collection[Path]: Path to the output images that have been convolved.
    """
    image_set = wsclean_result.image_set
    assert isinstance(image_set, ImageSet), (
        f"{image_set=}, but should be ImageSet instance"
    )
    return convolve_image_set(
        image_set=image_set,
        beam_shape=beam_shape,
        cutoff=cutoff,
        mode=mode,
        filter_str=filter_str,
        convol_suffix_str=convol_suffix_str,
        remove_original_images=remove_original_images,
    )


@task
def task_linmos_images(
    image_list: list[Path],
    container: Path,
    linmos_options: LinmosOptions,
    field_summary: FieldSummary | None = None,
    suffix_str: str | None = None,
    parset_output_path: str | None = None,
) -> LinmosResult:
    """Linmos together a set of input images.

    The ``linmos_options.image_output_name`` is updated to be
    a name based on the commonn unique fields of ``image_list``.

    Args:
        image_list (list[Path]): Collection of images to coadd
        container (Path): Path to a yandasoft singularity container
        linmos_options (LinmosOptions): Collection of linmos options
        field_summary (FieldSummary | None, optional): Description of the field, used to get the ``pol_axis`` of the field. Defaults to None.
        suffix_str (str | None, optional): Additional suffix str to add when generating the output file names. Defaults to None.
        parset_output_path (str | None, optional): The output parameter set that will be generated. Defaults to None.

    Returns:
        LinmosResult: Collection of linmos items generated
    """
    # TODO: Need a better filter_str approach. Would probably be better to
    # have literals for the type of product (MFS, cube, model) to be
    # sure of appropriate extraction

    from flint.naming import create_linmos_base_path
    from flint.utils import flatten_items

    image_list = flatten_items(items=image_list)

    output_path = create_linmos_base_path(
        input_images=image_list,
        additional_suffixes=suffix_str,
    )

    linmos_options = linmos_options.with_options(
        base_output_name=output_path,
        pol_axis=field_summary.pol_axis if field_summary else None,
    )

    linmos_result = linmos_images(
        images=image_list,
        linmos_options=linmos_options,
        container=container,
    )

    return linmos_result


def convolve_then_linmos(
    wsclean_results: Collection[WSCleanResult],
    beam_shape: BeamShape,
    field_options: FieldOptions | SubtractFieldOptions,
    linmos_suffix_str: str | None,
    field_summary: FieldSummary | None = None,
    convol_mode: str = "image",
    convol_filter: str = ".MFS.",
    convol_suffix_str: str = "conv",
    trim_linmos_fits: bool = True,
    remove_original_images: bool = False,
    cleanup_linmos: bool = False,
) -> LinmosResult:
    """An internal function that launches the convolution to a common resolution
    and subsequent linmos of the wsclean residual images.

    Args:
        wsclean_results (Collection[WSCleanResult]): Collection of wsclean imaging results, with residual images described in the attached ``ImageSet``
        beam_shape (BeamShape): The beam shape that residual images will be convolved to
        field_options (FieldOptions): Options related to the processing of the field
        linmos_suffix_str (str): The suffix string passed to the linmos parset name
        field_summary (Optional[FieldSummary], optional): The summary of the field, including (importantly) to orientation of the third-axis. Defaults to None.
        convol_mode (str, optional): The mode passed to the convol task to describe the images to extract. Support image or residual.  Defaults to image.
        convol_filter (str, optional): A text file applied when assessing images to co-add. Defaults to '.MFS.'.
        convol_suffix_str (str, optional): The suffix added to the convolved images. Defaults to 'conv'.
        trim_linmos_fits (bool, optional): Attempt to trim the output linmos files of as much empty space as possible. Defaults to True.
        remove_original_images (bool, optional): If True remove the original image after they have been convolved. Defaults to False.
        cleanup_linmos (bool, optional): Clean up items created throughout linmos, including the per-channel weight text files for each input image. Defaults to False.

    Returns:
        LinmosResult: Resulting linmos command parset
    """

    conv_images = task_convolve_image.map(
        wsclean_result=wsclean_results,
        beam_shape=unmapped(beam_shape),  # type: ignore
        cutoff=field_options.beam_cutoff,
        mode=convol_mode,
        filter_str=convol_filter,
        convol_suffix_str=convol_suffix_str,
        remove_original_images=remove_original_images,
    )
    assert field_options.yandasoft_container is not None
    # Though pol axis could be obtained from field_summary, at this point
    # it could be a PrefectFuture. Pass it over to the task.
    linmos_options = LinmosOptions(
        holofile=field_options.holofile,
        cutoff=field_options.pb_cutoff,
        trim_linmos_fits=trim_linmos_fits,
        cleanup=cleanup_linmos,
        remove_original_images=remove_original_images,
    )
    parset = task_linmos_images.submit(
        image_list=flatten_items(items=conv_images),
        container=field_options.yandasoft_container,
        suffix_str=linmos_suffix_str,
        linmos_options=linmos_options,
        field_summary=field_summary,
    )  # type: ignore

    return parset


def create_convol_linmos_images(
    wsclean_results: Collection[WSCleanResult],
    field_options: FieldOptions,
    field_summary: FieldSummary | None = None,
    additional_linmos_suffix_str: str | None = None,
) -> list[LinmosResult]:
    """Derive the appropriate set of beam shapes and then produce corresponding
    convolved and co-added images

    Args:
        wsclean_results (Collection[WSCleanResult]): Set of wsclean commands that have been executed
        field_options (FieldOptions): Set of field imaging options, containing details of the beam/s
        field_summary (Optional[FieldSummary], optional): Summary of the MSs, importantly containing their third-axis rotation. Defaults to None.
        additional_linmos_suffix_str (Optional[str], optional): An additional string added to the end of the auto-generated linmos base name. Defaults to None.

    Returns:
        List[LinmosResult]: The collection of linmos commands executed.
    """
    parsets: list[LinmosResult] = []

    beam_str: str = get_beam_resolution_str(mode="optimal")
    linmos_suffixes: list[str] = [beam_str]
    if additional_linmos_suffix_str:
        linmos_suffixes.insert(0, additional_linmos_suffix_str)

    # set up the expected name formats
    linmos_suffix_str = ".".join(linmos_suffixes)
    convol_suffix_str = f"{beam_str}.conv"

    beam_shape = task_get_common_beam_from_results.submit(
        wsclean_results=wsclean_results,
        cutoff=field_options.beam_cutoff,
        filter_str=".MFS.",
    )
    # NOTE: The order matters here. The last linmos file is used
    # when running the source finding. Putting this order around means
    # we would source find on the residual image
    if field_options.linmos_residuals:
        parsets.append(
            convolve_then_linmos(
                wsclean_results=wsclean_results,
                beam_shape=beam_shape,  # type: ignore
                field_options=field_options,
                linmos_suffix_str=f"{linmos_suffix_str}.residual",
                field_summary=field_summary,
                convol_mode="residual",
                convol_filter=".MFS.",
                convol_suffix_str=convol_suffix_str,
            )
        )
    parsets.append(
        convolve_then_linmos(
            wsclean_results=wsclean_results,
            beam_shape=beam_shape,  # type: ignore
            field_options=field_options,
            linmos_suffix_str=linmos_suffix_str,
            field_summary=field_summary,
            convol_mode="image",
            convol_filter=".MFS.",
            convol_suffix_str=convol_suffix_str,
        )
    )

    return parsets


@task
def task_convolve_linmos_to_fixed_shape(
    linmos_result: LinmosResult, field_options: FieldOptions
) -> LinmosResult:
    """Smooth the linmos input image to a fixed resolution

    Args:
        linmos_result (LinmosResult): An existing linmos results set representing a created image
        field_options (FieldOptions): A field options instance with an set fixed beeam shape set

    Returns:
        LinmosResult: A smoothed version of the input linmos image
    """
    from flint.naming import update_beam_resolution_field_in_path

    image_to_smooth = linmos_result.image_fits

    output_image_path: Path = update_beam_resolution_field_in_path(
        path=image_to_smooth, original_mode="optimal", updated_mode="fixed", marker="."
    )

    assert field_options.fixed_beam_shape, (
        f"{field_options.fixed_beam_shape=}, which is not allowed"
    )
    beam_shape = BeamShape(
        bmaj_arcsec=field_options.fixed_beam_shape[0],
        bmin_arcsec=field_options.fixed_beam_shape[1],
        bpa_deg=field_options.fixed_beam_shape[2],
    )

    # There is no option to provide a fully qualified name. So, capture then rename
    # the smoothed file
    images_to_smooth = [image_to_smooth]
    smoothed_linmos_paths = convolve_images(
        image_paths=images_to_smooth,
        beam_shape=beam_shape,
        convol_suffix="conv_to_rename",
        output_paths=[output_image_path],
    )
    assert len(smoothed_linmos_paths) == len(images_to_smooth) == 1, (
        f"Need matching path lengths between {images_to_smooth=} and {smoothed_linmos_paths=}"
    )

    return linmos_result.with_options(image_fits=output_image_path)


def create_convolve_linmos_cubes(
    wsclean_results: Collection[WSCleanResult],
    field_options: FieldOptions,
    current_round: int | None = None,
    additional_linmos_suffix_str: str | None = "cube",
):
    suffixes = [f"round{current_round}" if current_round is not None else "noselfcal"]
    if additional_linmos_suffix_str:
        suffixes.insert(0, additional_linmos_suffix_str)
    linmos_suffix_str = ".".join(suffixes)

    beam_shapes = task_get_cube_common_beam.submit(
        wsclean_results=wsclean_results, cutoff=field_options.beam_cutoff
    )
    convolved_cubes = task_convolve_cube.map(
        wsclean_result=wsclean_results,  # type: ignore
        cutoff=field_options.beam_cutoff,
        mode=unmapped("image"),  # type: ignore
        beam_shapes=unmapped(beam_shapes),  # type: ignore
    )

    assert field_options.yandasoft_container is not None
    linmos_options = LinmosOptions(
        holofile=field_options.holofile,
        cutoff=field_options.pb_cutoff,
    )
    parset = task_linmos_images.submit(
        image_list=convolved_cubes,  # type: ignore
        container=field_options.yandasoft_container,
        suffix_str=linmos_suffix_str,
        linmos_options=linmos_options,
    )
    return parset


@task
def task_create_image_mask_model(
    image: LinmosResult | ImageSet | WSCleanResult,
    image_products: AegeanOutputs,
    update_masking_options: dict[str, Any] | None = None,
) -> FITSMaskNames:
    """Create a mask for an image, with the intention of providing it as a clean mask
    to an appropriate imager.

    Args:
        linmos_parset (LinmosResult): Linmos command and associated meta-data
        image_products (AegeanOutputs): Images of the RMS and BKG
        update_masking_options (Optional[Dict[str,Any]], optional): Updated options supplied to the default MaskingOptions. Defaults to None.


    Raises:
        ValueError: Raised when ``image_products`` are not known

    Returns:
        FITSMaskNames: Clean mask where all pixels below a S/N are masked
    """
    if isinstance(image_products, AegeanOutputs):
        source_rms = image_products.rms
        source_bkg = image_products.bkg
    else:
        raise ValueError("Unsupported bkg/rms mode. ")

    source_image = None
    if isinstance(image, LinmosResult):
        source_image = image.image_fits
    elif isinstance(image, ImageSet) and image.image is not None:
        source_image = list(image.image)[-1]
    elif isinstance(image, WSCleanResult) and image.image_set is not None:
        source_image = list(image.image_set.image)[-1]
    else:
        source_image = image_products.image

    if source_image is None:
        raise ValueError(f"Unsupported image mode. Received {type(image)} ")

    logger.info(f"Creating a clean mask for {source_image=}")
    logger.info(f"Using {source_rms=}")
    logger.info(f"Using {source_bkg=}")

    masking_options = MaskingOptions()
    if update_masking_options:
        masking_options = masking_options.with_options(**update_masking_options)

    mask_names = create_snr_mask_from_fits(
        fits_image_path=source_image,
        fits_bkg_path=source_bkg,
        fits_rms_path=source_rms,
        masking_options=masking_options,
        create_signal_fits=True,
    )

    logger.info(f"Created {mask_names.mask_fits}")

    return mask_names


@task
def task_extract_beam_mask_image(
    linmos_mask_names: FITSMaskNames, wsclean_result: WSCleanResult
) -> FITSMaskNames:
    """Extract a clean mask for a beam from a larger clean mask (e.g. derived from a field)

    Args:
        linmos_mask_names (FITSMaskNames): Mask that will be drawn from to form a smaller clean mask (e.g. for a beam)
        wsclean_result (WSCleanResult): Wsclean command and meta-data. This is used to draw from the WCS to create an appropriate pixel-to-pixel mask

    Returns:
        FITSMaskNames: Clean mask for a image
    """
    # All images made by wsclean will have the same WCS
    assert wsclean_result.image_set is not None, (
        f"{wsclean_result.image_set=}, which should not happen"
    )
    beam_image = next(iter(wsclean_result.image_set.image))
    beam_mask_names = extract_beam_mask_from_mosaic(
        fits_beam_image_path=beam_image, fits_mosaic_mask_names=linmos_mask_names
    )

    return beam_mask_names


@task
def task_create_validation_plot(
    field_summary: FieldSummary,
    aegean_outputs: AegeanOutputs,
    reference_catalogue_directory: Path,
    upload_artifact: bool = True,
) -> Path:
    """Create a multi-panel figure highlighting the RMS, flux scale and astrometry of a field

    Args:
        aegean_outputs (AegeanOutputs): Output aegean products
        reference_catalogue_directory (Path): Directory containing NVSS, SUMSS and ICRS reference catalogues. These catalogues are reconginised internally and have expected names.
        upload_artifact (bool, optional): If True the validation plot will be uploaded to the prefect service as an artifact. Defaults to True.

    Returns:
        Path: Path to the output figure created
    """
    output_path = aegean_outputs.comp.parent

    logger.info(f"Will create validation plot in {output_path=}")

    plot_path = create_validation_plot(
        field_summary=field_summary,
        rms_image_path=aegean_outputs.rms,
        source_catalogue_path=aegean_outputs.comp,
        output_path=output_path,
        reference_catalogue_directory=reference_catalogue_directory,
    )

    if upload_artifact:
        upload_image_as_artifact(
            image_path=plot_path, description=f"Validation plot {plot_path!s}"
        )

    return plot_path


@task
def task_create_validation_tables(
    field_summary: FieldSummary,
    aegean_outputs: AegeanOutputs,
    reference_catalogue_directory: Path,
    upload_artifacts: bool = True,
) -> ValidationTables:
    """Create a set of validation tables that can be used to assess the
    correctness of an image and associated source catalogue.

    Args:
        processed_ms_paths (List[Path]): The processed MS files that were used to create the source catalogue
        rms_image_path (Path): The RMS fits image the source catalogue was constructed against.
        source_catalogue_path (Path): The source catalogue.
        output_path (Path): The output path of the figure to create
        reference_catalogue_directory (Path): The directory that contains the reference catalogues installed

    Returns:
        ValidationTables: The tables that were created
    """
    output_path = aegean_outputs.comp.parent

    logger.info(f"Will create validation tables in {output_path=}")
    validation_tables = create_validation_tables(
        field_summary=field_summary,
        rms_image_path=aegean_outputs.rms,
        source_catalogue_path=aegean_outputs.comp,
        output_path=output_path,
        reference_catalogue_directory=reference_catalogue_directory,
    )

    if upload_artifacts:
        for table in validation_tables:
            if isinstance(table, Path):
                validation_df = pd.read_csv(table)
                df_dict = validation_df.to_dict("records")
                create_table_artifact(
                    table=df_dict,
                    description=f"{table.stem}",
                )  # type: ignore
            elif isinstance(table, XMatchTables):
                for subtable in table:
                    if subtable is None:
                        continue
                    if not isinstance(subtable, Path):
                        continue
                    sub_df = pd.read_csv(subtable)
                    df_dict = sub_df.to_dict("records")
                    create_table_artifact(
                        table=df_dict,
                        description=f"{subtable.stem}",
                    )  # type: ignore

    return validation_tables


def validation_items(
    field_summary: FieldSummary,
    aegean_outputs: AegeanOutputs,
    reference_catalogue_directory: Path,
):
    """Construct the validation plot and validation table items for the imaged field.

    Internally these are submitting the prefect task versions of:
    - `task_create_validation_plot`
    - `task_create_validation_tables`

    Args:
        field_summary (FieldSummary): Container representing the SBID being imaged and its populated characteristics
        aegean_outputs (AegeanOutputs): Source finding results
        reference_catalogue_directory (Path): Location of directory containing the reference known NVSS, SUMSS and ICRS catalogues
    """

    val_plot_path = task_create_validation_plot.submit(
        field_summary=field_summary,
        aegean_outputs=aegean_outputs,
        reference_catalogue_directory=reference_catalogue_directory,
    )
    validate_tables = task_create_validation_tables.submit(
        field_summary=field_summary,
        aegean_outputs=aegean_outputs,
        reference_catalogue_directory=reference_catalogue_directory,
    )

    return (val_plot_path, validate_tables)



"""Itemss around testing components of bptools"""

from __future__ import annotations

import numpy as np
import pytest

from flint.bptools.preflagger import (
    construct_jones_over_max_amp_flags,
    construct_mesh_ant_flags,
)


def count_nan(data):
    return np.sum(~np.isfinite(data))


def test_construct_mesh_ant_flags():
    shape = (1, 36, 288, 4)
    a = np.arange(np.prod(shape)).reshape(shape) * 1.0

    assert count_nan(a) == 0

    a[0, 0, 10:20, :] = np.nan
    assert count_nan(a) == 40

    mask = construct_mesh_ant_flags(mask=~np.isfinite(a[0]))
    assert np.sum(mask) == 1440

    a[0, mask] = np.nan
    assert count_nan(a) == 1440


def test_construct_mesh_ant_flags_assert():
    shape = (1, 36, 288, 4)
    a = np.arange(np.prod(shape)).reshape(shape) * 1.0

    with pytest.raises(AssertionError):
        construct_mesh_ant_flags(mask=~np.isfinite(a))


def test_construct_jobnes_over_max_amp_flags():
    shape = (1, 36, 288, 4)
    a = np.arange(np.prod(shape)).reshape(shape) * 1.0

    a[0, 10, 10, 0] = 100000000
    max_amplitude = 100000
    assert np.sum(a > max_amplitude) == 1

    mask = construct_jones_over_max_amp_flags(
        complex_gains=a, max_amplitude=max_amplitude
    )
    assert np.sum(mask) == 4

    a[mask] = np.nan
    assert count_nan(a) == 4


def test_construct_jobnes_over_max_amp_flags2():
    shape = (1, 36, 288, 4)
    a = np.arange(np.prod(shape)).reshape(shape) * 1.0

    a[0, 10, 10, 0] = 100000000
    a[0, 10, 11, 1] = 100000000
    a[0, 11, 12, 2] = 100000000
    a[0, 12, 13, 3] = 100000000
    max_amplitude = 100000
    assert np.sum(a > max_amplitude) == 4

    mask = construct_jones_over_max_amp_flags(
        complex_gains=a, max_amplitude=max_amplitude
    )
    assert np.sum(mask) == 16

    a[mask] = np.nan
    assert count_nan(a) == 16


def test_construct_jobnes_over_max_amp_flags3():
    shape = (1, 36, 288, 5)
    a = np.arange(np.prod(shape)).reshape(shape) * 1.0

    with pytest.raises(AssertionError):
        construct_jones_over_max_amp_flags(complex_gains=a, max_amplitude=1000000)



"""Contains the core of the option class containers that are used to
hold stateful properties throughout the flint codebase.
"""

# NOTE: Although these options could be stored closer to where
# their logic is often used, at times these can cause circular dependencies.
# This happens a lot with the linting / typing checking, where classes are
# imported purely for tools like ruff

from __future__ import (  # Used for mypy/pylance to like the return type of MS.with_options
    annotations,
)

from argparse import ArgumentParser, Namespace
from pathlib import Path
from types import NoneType, UnionType
from typing import (
    Any,
    NamedTuple,
    TypeVar,
    get_args,
    get_origin,
)

import yaml
from pydantic import BaseModel, ConfigDict
from pydantic.fields import FieldInfo

from flint.exceptions import MSError
from flint.logging import logger


class MS(NamedTuple):
    """Helper to keep track of measurement set information

    This is the class that should be used when describing a measurement
    set that will be operated on.
    """

    path: Path
    """Path to the measurement set that is being represented"""
    column: str | None = None
    """Column that should be operated against"""
    beam: int | None = None
    """The beam ID of the MS within an ASKAP field"""
    spw: int | None = None
    """Intended to be used with ASKAP high-frequency resolution modes, where the MS is divided into SPWs"""
    field: str | None = None
    """The field name  of the data"""
    model_column: str | None = None
    """The column name of the most recently MODEL data"""

    @property
    def ms(self) -> MS:
        return self

    @classmethod
    def cast(cls, ms: MS | Path) -> MS:
        """Create/return a MS instance given either a Path or MS.

        If the input is neither a MS instance or Path, the object will
        be checked to see if it has a `.ms` attribute. If it does then
        this will be used.

        Args:
            ms (Union[MS, Path]): The input type to consider

        Raises:
            MSError: Raised when the input ms can not be cast to an MS instance

        Returns:
            MS: A normalised MS
        """
        if isinstance(ms, MS):
            # Nothing to do
            pass
        elif isinstance(ms, Path):
            ms = MS(path=ms)
        elif "ms" in dir(ms) and isinstance(ms.ms, MS):
            ms = ms.ms
        else:
            raise MSError(f"Unable to convert {ms=} of {type(ms)} to MS object. ")

        return ms

    def with_options(self, **kwargs) -> MS:
        """Create a new MS instance with keywords updated

        Returns:
            MS: New MS instance with updated attributes
        """
        # TODO: Update the signature to have the actual attributes to
        # help keep mypy and other linters happy
        as_dict = self._asdict()
        as_dict.update(kwargs)

        return MS(**as_dict)


def options_to_dict(input_options: Any) -> dict:
    """Helper function to convert an `Options` type class to a dictionary.

    Most of `flint` `Option` and `Result` classes used `typing.NamedTuples`, which carry with
    it a `_asdict` method to convert them to a dictionary. Future roadmap plans to move over to
    pydantic type models. This is a place holder function to help transition to this.

    Args:
        input_options (Any): Item to convert to a dictionary

    Raises:
        TypeError: Raised if the conversion to a dictionary was not successful

    Returns:
        Dict: The dictionary version of the input options
    """

    if "_asdict" in dir(input_options):
        return input_options._asdict()

    try:
        if issubclass(input_options, BaseModel):
            return dict(**input_options.__dict__)
    except TypeError:
        logger.debug(f"can not use issubclass on {input_options}")

    try:
        return dict(**input_options)
    except TypeError:
        raise TypeError(f"Input options is not known: {type(input_options)}")


T = TypeVar("T", bound=BaseModel)


class BaseOptions(BaseModel):
    """A base class that Options style flint classes can
    inherit from. This is derived from ``pydantic.BaseModel``,
    and can be used for validation of supplied values.

    Class derived from ``BaseOptions`` are immutable by
    default, and have the docstrings of attributes
    extracted.
    """

    model_config = ConfigDict(
        frozen=True, from_attributes=True, use_attribute_docstrings=True, extra="forbid"
    )

    def with_options(self: T, /, **kwargs) -> T:
        new_args = self.__dict__.copy()
        new_args.update(**kwargs)

        return self.__class__(**new_args)

    def _asdict(self) -> dict[str, Any]:
        return self.__dict__


def _create_argparse_options(name: str, field: FieldInfo) -> tuple[str, dict[str, Any]]:
    """Convert a pydantic Field into ``dict`` to splate into ArgumentParser.add_argument()"""

    field_name = name if field.is_required() else "--" + name.replace("_", "-")

    field_type = get_origin(field.annotation)
    field_args = get_args(field.annotation)
    iterable_types = (list, tuple, set)

    options = dict(action="store", help=field.description, default=field.default)

    if field.annotation is bool:
        options["action"] = "store_false" if field.default else "store_true"

    # if field_type is in (list, tuple, set) OR if (list, tuple, set) | Any
    elif field_type in iterable_types or (
        field_type is UnionType
        and any(get_origin(p) in iterable_types for p in field_args)
    ):
        nargs: str | int = "+"

        # If the field is a tuple, and the Ellipsis is not present
        # We can assume that the nargs is the length of the tuple
        if field_type is tuple and Ellipsis not in field_args:
            nargs = len(field_args)

        # Now we handle unions, but do the same check as above
        elif field_type is UnionType and Ellipsis not in field_args:
            for arg in field_args:
                args = get_args(arg)
                if arg is not NoneType and type(args) is tuple and Ellipsis not in args:
                    nargs = len(args)

        if nargs == 0:
            raise ValueError(f"Unable to determine nargs for {name=}, got {nargs=}")
        options["nargs"] = nargs

    return field_name, options


def add_options_to_parser(
    parser: ArgumentParser,
    options_class: type[BaseOptions],
    description: str | None = None,
) -> ArgumentParser:
    """Given an established argument parser and a class derived
    from a ``pydantic.BaseModel``, populate the argument parser
    with the model properties.

    Args:
        parser (ArgumentParser): Parser that arguments will be added to
        options_class (type[BaseModel]): A ``Options`` style class derived from ``BaseOptions``

    Returns:
        ArgumentParser: Updated argument parser
    """

    assert issubclass(options_class, BaseModel), (
        f"{options_class=} is not a pydantic BaseModel"
    )

    group = parser.add_argument_group(
        title=f"Inputs for {options_class.__name__}", description=description
    )

    for name, field in options_class.model_fields.items():
        field_name, options = _create_argparse_options(name=name, field=field)
        try:
            group.add_argument(field_name, **options)  # type: ignore
        except Exception as e:
            logger.error(f"{field_name=} {options=}")
            raise e

    return parser


U = TypeVar("U", bound=BaseOptions)


def create_options_from_parser(
    parser_namespace: Namespace, options_class: type[U]
) -> U:
    """Given a ``BaseOptions`` derived class, extract the corresponding
    arguments from an ``argparse.nNamespace``. These options correspond to
    ones generated by ``add_options_to_parser``.

    Args:
        parser_namespace (Namespace): The argument parser corresponding to those in the ``BaseOptions`` class
        options_class (U): A ``BaseOptions`` derived class

    Returns:
        U: An populated options class with arguments drawn from CLI argument parser
    """
    assert issubclass(
        options_class,  # type: ignore
        BaseModel,
    ), f"{options_class=} is not a pydantic BaseModel"

    args = (
        vars(parser_namespace)
        if not isinstance(parser_namespace, dict)
        else parser_namespace
    )

    opts_dict = {}
    for name, field in options_class.model_fields.items():
        opts_dict[name] = args[name]

    return options_class(**opts_dict)


class BandpassOptions(BaseOptions):
    """Container that reoresents the flint related options that
    might be used throughout the processing of bandpass calibration
    data.

    In its present form this `BandpassOptions` class is not intended
    to contain properties of the data that arebeing processed, rather
    how these data will be processed.

    These settings are not meant to be adjustabled throughout
    a single bandpass pipeline run
    """

    flagger_container: Path | None = None
    """Path to the singularity aoflagger container"""
    calibrate_container: Path | None = None
    """Path to the singularity calibrate container"""
    expected_ms: int = 36
    """The expected number of measurement set files to find"""
    smooth_solutions: bool = False
    """Will activate the smoothing of the bandpass solutions"""
    smooth_window_size: int = 16
    """The width of the smoothing window used to smooth the bandpass solutions"""
    smooth_polynomial_order: int = 4
    """The polynomial order used by the Savgol filter when smoothing the bandpass solutions"""
    flag_calibrate_rounds: int = 3
    """The number of times the bandpass will be calibrated, flagged, then recalibrated"""
    minuv: float | None = None
    """The minimum baseline length, in meters, for data to be included in bandpass calibration stage"""
    preflagger_ant_mean_tolerance: float = 0.2
    """Tolerance that the mean x/y antenna gain ratio test before the antenna is flagged"""
    preflagger_mesh_ant_flags: bool = False
    """Share channel flags from bandpass solutions between all antenna"""
    preflagger_jones_max_amplitude: float | None = None
    """Flag Jones matrix if any amplitudes with a Jones are above this value"""


class AddModelSubtractFieldOptions(BaseOptions):
    """Options related to predicting a continuum model during the SubtractFieldOptions workflow.
    Specifically these options deal with identifying the wsclean produced source list model, which
    may be used by ``admodel`` to predict model visibilities. See utilities around the ``aocalibrate``
    functions and routines."""

    wsclean_pol_mode: list[str] = ["i"]
    """The polarisation of the wsclean model that was generated"""
    calibrate_container: Path | None = None
    """Path to the container with the calibrate software (including addmodel)"""
    addmodel_cluster_config: Path | None = None
    """Specify a new cluster configuration file different to the preferred on. If None, drawn from preferred cluster config"""


class SubtractFieldOptions(BaseOptions):
    """Container for options related to the
    continuum-subtracted pipeline"""

    wsclean_container: Path
    """Path to the container with wsclean"""
    yandasoft_container: Path
    """Path to the container with yandasoft"""
    subtract_model_data: bool = False
    """Subtract the MODEL_DATA column from the nominated data column"""
    data_column: str = "CORRECTED_DATA"
    """Describe the column that should be imaed and, if requested, have model subtracted from"""
    expected_ms: int = 36
    """The number of measurement sets that should exist"""
    imaging_strategy: Path | None = None
    """Path to a FLINT imaging yaml file that contains settings to use throughout imaging"""
    holofile: Path | None = None
    """Path to the holography FITS cube that will be used when co-adding beams"""
    linmos_residuals: bool = False
    """Linmos the cleaning residuals together into a field image"""
    beam_cutoff: float = 150
    """Cutoff in arcseconds to use when calculating the common beam to convol to"""
    pb_cutoff: float = 0.1
    """Primary beam attenuation cutoff to use during linmos"""
    stagger_delay_seconds: float | None = None
    """The delay, in seconds, that should be used when submitting items in batches (e.g. looping over channels)"""
    attempt_subtract: bool = False
    """Attempt to subtract the model column from the nominated data column"""
    subtract_data_column: str = "DATA"
    """Should the continuum model be subtracted, where to store the output"""
    predict_wsclean_model: bool = False
    """Search for the continuum model produced by wsclean and subtract"""
    use_addmodel: bool = False
    """Invoke the ``addmodel`` visibility prediction, including the search for the ``wsclean`` source list"""
    use_crystalball: bool = False
    """Attempt to predict the model visibilities using ``crystalball``"""


class FieldOptions(BaseOptions):
    """Container that represents the flint related options that
    might be used throughout components related to the actual
    pipeline.

    In its present form this `FieldOptions` class is not intended
    to contain properties of the data that are being processed,
    rather how those data will be processed.

    These settings are not meant to be adjustable throughout
    rounds of self-calibration.
    """

    flagger_container: Path | None = None
    """Path to the singularity aoflagger container"""
    calibrate_container: Path | None = None
    """Path to the singularity calibrate container"""
    casa_container: Path | None = None
    """Path to the singularity CASA container"""
    expected_ms: int = 36
    """The expected number of measurement set files to find"""
    wsclean_container: Path | None = None
    """Path to the singularity wsclean container"""
    yandasoft_container: Path | None = None
    """Path to the singularity yandasoft container"""
    potato_container: Path | None = None
    """Path to the singularity potato peel container"""
    holofile: Path | None = None
    """Path to the holography FITS cube that will be used when co-adding beams"""
    rounds: int = 2
    """Number of required rouds of self-calibration and imaging to perform"""
    skip_selfcal_on_rounds: list[int] | None = None
    """Do not perform the derive and apply self-calibration solutions on these rounds"""
    zip_ms: bool = False
    """Whether to zip measurement sets once they are no longer required"""
    run_aegean: bool = False
    """Whether to run the aegean source finding tool"""
    aegean_container: Path | None = None
    """Path to the singularity aegean container"""
    no_imaging: bool = False
    """Whether to skip the imaging process (including self-calibration)"""
    reference_catalogue_directory: Path | None = None
    """Path to the directory container the reference catalogues, used to generate validation plots"""
    linmos_residuals: bool = False
    """Linmos the cleaning residuals together into a field image"""
    beam_cutoff: float = 150
    """Cutoff in arcseconds to use when calculating the common beam to convol to"""
    fixed_beam_shape: tuple[float, float, float] | None = None
    """Specify the final beamsize of linmos field images in (arcsec, arcsec, deg)"""
    pb_cutoff: float = 0.1
    """Primary beam attenuation cutoff to use during linmos"""
    use_preflagger: bool = False
    """Whether to apply (or search for solutions with) bandpass solutions that have gone through the preflagging operations"""
    use_smoothed: bool = False
    """Whether to apply (or search for solutions with) a bandpass smoothing operation applied"""
    use_beam_masks: bool = False
    """Construct beam masks from MFS images to use for the next round of imaging. """
    use_beam_masks_from: int = 1
    """If `use_beam_masks` is True, this sets the round where beam masks will be generated from"""
    use_beam_masks_rounds: list[int] | None = None
    """If `use_beam_masks` is True, this sets which rounds should have a mask applied"""
    imaging_strategy: Path | None = None
    """Path to a FLINT imaging yaml file that contains settings to use throughout imaging"""
    sbid_archive_path: Path | None = None
    """Path that SBID archive tarballs will be created under. If None no archive tarballs are created. See ArchiveOptions. """
    sbid_copy_path: Path | None = None
    """Path that final processed products will be copied into. If None no copying of file products is performed. See ArchiveOptions. """
    rename_ms: bool = False
    """Rename MSs throughout rounds of imaging and self-cal instead of creating copies. This will delete data-columns throughout. """
    stokes_v_imaging: bool = False
    """Specifies whether Stokes-V imaging will be carried out after the final round of imagine (whether or not self-calibration is enabled). """
    coadd_cubes: bool = False
    """Co-add cubes formed throughout imaging together. Cubes will be smoothed channel-wise to a common resolution. Only performed on final set of images"""
    update_model_data_with_source_list: bool = False
    """Attempt to update a MSs MODEL_DATA column with a source list (e.g. source list output from wsclean)"""


class PolFieldOptions(BaseOptions):
    """Container that represents the flint related options that
    might be used throughout components related to the actual
    pipeline.

    In its present form this `PolFieldOptions` class is not intended
    to contain properties of the data that are being processed,
    rather how those data will be processed.

    These settings are not meant to be adjustable across different polarisations.
    """

    expected_ms: int = 36
    """The expected number of measurement set files to find"""
    wsclean_container: Path | None = None
    """Path to the singularity wsclean container"""
    yandasoft_container: Path | None = None
    """Path to the singularity yandasoft container"""
    casa_container: Path | None = None
    """Path to the singularity CASA container"""
    holofile: Path | None = None
    """Path to the holography FITS cube that will be used when co-adding beams"""
    beam_cutoff: float = 150
    """Cutoff in arcseconds to use when calculating the common beam to convol to"""
    fixed_beam_shape: tuple[float, float, float] | None = None
    """Specify the final beamsize of linmos field images in (arcsec, arcsec, deg)"""
    pb_cutoff: float = 0.1
    """Primary beam attenuation cutoff to use during linmos"""
    trim_linmos_fits: bool = False
    """Trim the linmos fits files to remove the padding that is added. If True, the output fits files will be smaller but might be different shapes"""
    imaging_strategy: Path | None = None
    """Path to a FLINT imaging yaml file that contains settings to use throughout imaging"""
    sbid_copy_path: Path | None = None
    """Path that final processed products will be copied into. If None no copying of file products is performed. See ArchiveOptions. """


def dump_field_options_to_yaml(
    output_path: Path,
    field_options: FieldOptions | PolFieldOptions | SubtractFieldOptions,
    overwrite: bool = False,
) -> Path:
    """Dump the supplied instance of `FieldOptions` to a yaml file
    for record keeping.

    The parent directory of the `output_path` will be created if it
    does not already exist.

    Args:
        output_path (Path): Path of the output file.
        field_options (FieldOptions): The `FieldOptions` class to write.
        overwrite (bool, optional): Overwrite the file if it exists. Defaults to False.

    Raises:
        FileExistsError: Raise if `output_path` already exists and `overwrite` is `False`

    Returns:
        Path: Output path written to.
    """

    logger.info(f"Writing field_options to {output_path}")

    if not overwrite and output_path.exists():
        raise FileExistsError(f"{output_path=} exists. ")

    # Create the directory just in case
    output_path.parent.mkdir(parents=True, exist_ok=True)

    with open(output_path, "w") as out_file:
        yaml.dump(data=field_options._asdict(), stream=out_file, sort_keys=False)

    return output_path


# TODO: Perhaps move these to flint.naming, and can be built up
# based on rules, e.g. imager used, source finder etc.
DEFAULT_TAR_RE_PATTERNS = (
    r".*MFS.*image\.fits",
    r".*linmos.*",
    r".*weight\.fits",
    r".*yaml",
    r".*\.txt",
    r".*png",
    r".*beam[0-9]+\.ms\.zip",
    r".*beam[0-9]+\.ms",
    r".*\.caltable",
    r".*\.tar",
    r".*\.csv",
)
DEFAULT_COPY_RE_PATTERNS = (r".*linmos.*fits", r".*weight\.fits", r".*png", r".*csv")


class ArchiveOptions(BaseOptions):
    """Container for options related to archiving products from flint workflows"""

    tar_file_re_patterns: tuple[str, ...] = DEFAULT_TAR_RE_PATTERNS
    """Regular-expressions to use to collect files that should be tarballed"""
    copy_file_re_patterns: tuple[str, ...] = DEFAULT_COPY_RE_PATTERNS
    """Regular-expressions used to identify files to copy into a final location (not tarred)"""



"""Attempts to centralise components to do with naming of pipeline files and data
products.
"""

from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path
from typing import Any, Literal, NamedTuple, TypeVar

from flint.exceptions import NamingException
from flint.logging import logger

PathStr = TypeVar("PathStr", str, Path)


def _rename_linear_to_stokes(
    linear_name_str: str,
    stokes: str,
) -> str:
    if stokes.lower() not in ("q", "u"):
        raise NameError(f"Stokes {stokes=} is not linear!")
    pattern = r"\.qu"  # Regex pattern to replace
    stokes_name = re.sub(pattern, f".{stokes}", linear_name_str)
    logger.info(f"Renamed {linear_name_str=} to {stokes_name=}")
    return stokes_name


def rename_linear_to_stokes(
    linear_name: PathStr,
    stokes: str,
) -> PathStr:
    if isinstance(linear_name, Path):
        return Path(_rename_linear_to_stokes(linear_name.as_posix(), stokes))

    return _rename_linear_to_stokes(linear_name, stokes)


def get_fits_cube_from_paths(paths: list[Path]) -> list[Path]:
    """Given a list of files, find the ones that appear to be FITS files
    and contain the ``.cube.`` field indicator. A regular expression searching
    for both the ``.cube.`` and ``.fits`` file type is used.

    Args:
        paths (List[Path]): The set of paths to examine to identify potential cube fits images from

    Returns:
        List[Path]: Set of paths matching the search criteria
    """
    cube_expression = re.compile(r"\.cube\..*fits$")

    cube_files = [path for path in paths if bool(cube_expression.search(str(path)))]

    return cube_files


LONG_FIELD_TO_SHORTHAND = {
    "sbid": "SB",
    "beam": "beam",
    "channel_range": "ch",
    "round": "round",
}
"""Name mapping between the longform of ProcessedFieldComponents and shorthands used"""


def create_name_from_common_fields(
    in_paths: tuple[Path, ...], additional_suffixes: str | None = None
) -> Path:
    """Attempt to craft a base name using the field elements that are in common.
    The expectation that these are paths that can be processed by the ``processed_name_format``
    handler. Resulting fields that are common across all ``in_paths`` are preserved.

    Only fields that are recognised as a known property are retained. Suffixes that do not
    form a component are ignored. For example:

    >>> "59058/SB59058.RACS_1626-84.ch0287-0288.linmos.fits"

    the `linmos.fits` would be ignored.

    All ``in_paths`` should be detected, otherwise an ValueError is raised

    Args:
        in_paths (Tuple[Path, ...]): Collection of input paths to consider
        additional_suffixes (Optional[str], optional): Add an additional set of suffixes before returning. Defaults to None.

    Raises:
        ValueError: Raised if any of the ``in_paths`` fail to conform to ``flint`` processed name format

    Returns:
        Path: Common fields with the same base parent path
    """
    from flint.options import options_to_dict

    in_paths = tuple(Path(p) for p in in_paths)
    parent = in_paths[0].parent
    processed_components = list(map(processed_ms_format, in_paths))

    if None in processed_components:
        raise ValueError("Processed name format failed")
    processed_components_dict = [options_to_dict(pc) for pc in processed_components]

    keys_to_test = processed_components_dict[0].keys()
    logger.info(f"{keys_to_test=}")
    # Extract the fields that are constant across all inputs and are not None
    constant_fields = {
        key: processed_components_dict[0][key]
        for key in keys_to_test
        if len(set([pcd[key] for pcd in processed_components_dict])) == 1
        and processed_components_dict[0][key] is not None
    }

    name_path = create_path_from_processed_name_components(
        processed_name_components=ProcessedNameComponents(**constant_fields),
        parent_path=parent,
    )
    constant_field_keys = list(constant_fields.keys())
    logger.info(f"Identified {constant_field_keys=}")

    if additional_suffixes:
        additional_suffixes = (
            f".{additional_suffixes}"
            if not additional_suffixes.startswith(".")
            else additional_suffixes
        )
        name_path = Path(str(name_path) + additional_suffixes)
    return Path(name_path)


# TODO: Need to assess the mode argument, and define literals that are accepted
def create_image_cube_name(
    image_prefix: Path,
    mode: str | list[str] | None = None,
    suffix: str | list[str] | None = None,
) -> Path:
    """Create a consistent naming scheme when combining images into cube images. Intended to
    be used when combining many subband images together into a single cube.

    The name returned will be:
    >>> {image_prefix}.{mode}.{suffix}.cube.fits

    Should ``mode`` or ``suffix`` be a list, they will be joined with '.' separators. Hence, no
    '.' should be added.

    This function will always output 'cube.fits' at the end of the returned file name.

    Args:
        image_prefix (Path): The unique path of the name. Generally this is the common part among the input planes
        mode (Optional[Union[str, List[str]]], optional): Additional mode/s to add to the file name. Defaults to None.
        suffix (Optional[Union[str, List[str]]], optional): Additional suffix/s to add before the final 'cube.fits'. Defaults to None.

    Returns:
        Path: The final path and file name
    """
    # NOTE: This is likely a function to grow in time as more imaging and pipeline modes added. Putting
    # it here for future proofing
    output_cube_name = f"{Path(image_prefix)!s}.{mode}.{suffix}"

    output_components = [str(Path(image_prefix))]
    if mode:
        # TODO: Assess what modes are actually allowed. Suggestion is to
        # make a class of some sort with specified and known markers that
        # are opted into. Hate this "everything and anything"
        (
            output_components.append(mode)
            if isinstance(mode, str)
            else output_components.extend(mode)
        )
    if suffix:
        # TODO: See above. Need a class of acceptable suffixes to use
        (
            output_components.append(suffix)
            if isinstance(suffix, str)
            else output_components.extend(suffix)
        )

    output_components.append("cube.fits")

    output_cube_name = ".".join(output_components)
    return Path(output_cube_name)


def create_imaging_name_prefix(
    ms_path: Path,
    pol: str | None = None,
    channel_range: tuple[int, int] | None = None,
) -> str:
    """Given a measurement set and a polarisation, create the naming prefix to be used
    by some imager

    Args:
        ms (Union[MS,Path]): The measurement set being considered
        pol (Optional[str], optional): Whether a polarsation is being considered. Defaults to None.
        channel_range (Optional[Tuple[int,int]], optional): The channel range that is going to be imaged. Defaults to none.

    Returns:
        str: The constructed string name
    """

    names = [ms_path.stem]
    if pol:
        names.append(f"{pol.lower()}")
    if channel_range:
        names.append(f"ch{channel_range[0]:04}-{channel_range[1]:04}")

    return ".".join(names)


ResolutionModes = Literal["optimal", "fixed"]


def get_beam_resolution_str(mode: ResolutionModes, marker: str | None = None) -> str:
    """Map a beam resolution mode to an appropriate suffix. This
    is located her in anticipation of other imaging modes.

    Supported modes are: 'optimal', 'fixed', 'raw'

    Args:
        mode (Literal["fixed","optimal"]): The mode of image resolution to use.
        marker (Optional[str], optional): Append the marker to the end of the returned mode string. If None mode string is returned. Defaults to None.

    Raises:
        ValueError: Raised when an unrecognised mode is supplied

    Returns:
        str: The appropriate string for mapped mode
    """
    # NOTE: Arguably this is a trash and needless function. Adding it
    # in case other modes are ever needed or referenced. No idea whether
    # it will ever been needed and could be removed in future.
    supported_modes: dict[str, str] = dict(optimal="optimal", fixed="fixed", raw="raw")
    if mode.lower() not in supported_modes.keys():
        raise ValueError(
            f"Received {mode=}, supported modes are {supported_modes.keys()}"
        )

    mode_str = supported_modes[mode.lower()]

    return mode_str + marker if marker else mode_str


def update_beam_resolution_field_in_path(
    path: Path,
    original_mode: ResolutionModes,
    updated_mode: ResolutionModes,
    marker: str | None = None,
) -> Path:
    """Transition the resolution indicator in a processed name (either ``optimal`` or ``fixed``)
    to another state. For example:

    >>> 'SB57516.RACS_0929-81.round4.i.optimal.round4.residual.linmos.fits'

    to

    >>> 'SB57516.RACS_0929-81.round4.i.fixed.round4.residual.linmos.fits'

    See ``get_beam_resolution_str`` for addition information. Supported modes are
    ``fixed`` and ``optimal``

    Args:
        path (Path): The path to inspect and update
        original_mode (ResolutionModes): The original mode
        updated_mode (ResolutionModes): The mode to move to
        marker (str | None, optional): The marker to separate the field. Defaults to None.

    Returns:
        Path: Updated path
    """
    original_mode_str = get_beam_resolution_str(mode=original_mode, marker=marker)
    updated_mode_str = get_beam_resolution_str(mode=updated_mode, marker=marker)

    assert original_mode_str in str(path), f"{original_mode_str=} not in {path=}"
    new_path = Path(str(path).replace(original_mode_str, updated_mode_str))
    logger.info(
        f"Updated beam resolution mode from {original_mode=} to {updated_mode=}"
    )

    return new_path


def get_selfcal_ms_name(in_ms_path: Path, round: int = 1) -> Path:
    """Create the new output MS path that will be used for self-calibration. The
    output measurement set path will include a roundN.ms suffix, where N is the
    round. If such a suffix already exists from an earlier self-calibration round,
    it will be removed and replaced.

    Args:
        in_ms_path (Path): The measurement set that will go through self-calibration
        round (int, optional): The self-calibration round number that is currently being used. Defaults to 1.

    Returns:
        Path: Output measurement set path to use
    """
    res = re.search("\\.round[0-9]+.ms", str(in_ms_path.name))
    if res:
        logger.info("Detected a previous round of self-calibration. ")
        span = res.span()
        name_str = str(in_ms_path.name)
        name = f"{name_str[: span[0]]}.round{round}.ms"
    else:
        name = f"{in_ms_path.stem!s}.round{round}.ms"
    out_ms_path = in_ms_path.parent / name

    assert in_ms_path != out_ms_path, (
        f"{in_ms_path=} and {out_ms_path=} match. Something went wrong when creating new self-cal name. "
    )

    return out_ms_path


def add_timestamp_to_path(
    input_path: Path | str, timestamp: datetime | None = None
) -> Path:
    """Add a timestamp to a input path, where the timestamp is the
    current data and time. The time will be added to the name component
    before the file suffix. If the name component of the `input_path`
    has multiple suffixes than the timestamp will be added before the last.

    Args:
        input_path (Union[Path, str]): Path that should have a timestamp added
        timestamp: (Optional[datetime], optional): The date-time to add. If None the current time is used. Defaults to None.
    Returns:
        Path: Updated path with a timestamp in the file name
    """
    input_path = Path(input_path)
    timestamp = timestamp if timestamp else datetime.now()

    time_str = timestamp.strftime("%Y%m%d-%H%M%S")
    new_name = f"{input_path.stem}-{time_str}{input_path.suffix}"
    output_path = input_path.with_name(new_name)

    return output_path


class CASDANameComponents(NamedTuple):
    """Container for the components of a CASDA MS. These are really those
    processed by the ASKAP pipeline"""

    sbid: int
    """The sbid of the observation"""
    field: str
    """The name of the field extracted"""
    beam: str
    """Beam number of the data"""
    spw: str | None = None
    """If multiple MS were written as the data were in a high-frequency resolution mode, which segment"""
    alias: str | None = None
    """Older ASKAP MSs could be packed with multiple fields. The ASKAP pipeline holds this field as an alias. They are now the same in almost all cases as the field. """
    format: str = "science"
    """What the format / type of the data the MS is. """


def casda_ms_format(in_name: str | Path) -> CASDANameComponents | None:
    """Break up a CASDA sty;e MS name (really the askap pipeline format) into its recognised parts.
    if a match fails a `None` is returned.

    Example of a CASDA style MS:

    - `scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms`

    Args:
        in_name (Union[str, Path]): The path to or name of the MS to consider

    Returns:
        Union[CASDANameComponents, None]: The returned components of the MS. If this fails a `None` is returned.
    """

    in_name = Path(in_name).name

    # An example
    # scienceData.RACS_1237+00.SB40470.RACS_1237+00.beam35_averaged_cal.leakage.ms

    logger.debug(f"Matching {in_name}")
    regex = re.compile(
        r"^(?P<format>(scienceData|1934))(\.(?P<alias>.*))?\.SB(?P<sbid>[0-9]+)(\.(?P<field>.*))?\.beam(?P<beam>[0-9]+).*ms"
    )
    results = regex.match(in_name)

    if results is None:
        logger.debug(f"No casda_ms_format results to {in_name} found")
        return None

    return CASDANameComponents(
        sbid=int(results["sbid"]),
        field=results["field"],
        beam=results["beam"],
        alias=results["alias"],
        format=results["format"],
    )


class RawNameComponents(NamedTuple):
    date: str
    """Date that the data were taken, of the form YYYY-MM-DD"""
    time: str
    """Time that the data were written"""
    beam: str
    """Beam number of the data"""
    spw: str | None = None
    """If multiple MS were written as the data were in a high-frequency resolution mode, which segment"""


def raw_ms_format(in_name: str) -> None | RawNameComponents:
    """The typical ASKAP measurement written to the ingest disks
    has the form:

    >>> 2022-04-14_100122_1.ms

    and in the case of a multiple beams written out (in high frequency resolution mode)

    >>> 2022-04-14_100122_1_1.ms

    This function will attempt to break it up into its main parts
    and return the mapping.

    Args:
        in_name (str): The name of a file, presumably a measurement set. The left-most part will be examined for to identify the raw ASKAP naming scheme.

    Returns:
        Union[None,Dict[str,str]]: None if the raw ASKAP measurement set naming scheme was not detected, otherwise a dictionary representing its parts.
    """

    logger.debug(f"Matching {in_name}")
    regex = re.compile(
        "^(?P<date>[0-9]{4}-[0-9]{1,2}-[0-9]{1,2})_(?P<time>[0-9]+)_(?P<beam>[0-9]+)(_(?P<spw>[0-9]+))*"
    )
    results = regex.match(in_name)

    if results is None:
        logger.debug(f"No raw_ms_format results to {in_name} found")
        return None

    groups = results.groupdict()

    logger.debug(f"Matched groups are: {groups}")

    return RawNameComponents(
        date=groups["date"], time=groups["time"], beam=groups["beam"], spw=groups["spw"]
    )


class ProcessedNameComponents(NamedTuple):
    """Container for a file name derived from a MS flint name. Generally of the
    form: SB.Field.Beam.Spw"""

    sbid: str
    """The sbid of the observation"""
    field: str
    """The name of the field extracted"""
    beam: str | None = None
    """The beam of the observation processed"""
    spw: str | None = None
    """The SPW of the observation. If there is only one spw this is None."""
    round: str | None = None
    """The self-calibration round detected. This might be represented as 'noselfcal' in some image products, e.g. linmos. """
    pol: str | None = None
    """The polarisation component, if it exists, in a filename. Examples are 'i','q','u','v'. Could be combinations in some cases depending on how it was created (e.g. based on wsclean pol option). """
    channel_range: tuple[int, int] | None = None
    """The channel range encoded in an file name. Generally are zero-padded, and are two fields of the form ch1234-1235, where the upper bound is exclusive. Defaults to none."""


def processed_ms_format(
    in_name: str | Path,
) -> ProcessedNameComponents | None:
    """Will take a formatted name (i.e. one derived from the flint.naming.create_ms_name)
    and attempt to extract its main components. This includes the SBID, field, beam and spw.

    Args:
        in_name (Union[str, Path]): The name that needs to be broken down into components

    Returns:
        Union[FormattedNameComponents,None': A structure container the sbid, field, beam and spw. None is returned if can not be parsed.
    """

    in_name = in_name.name if isinstance(in_name, Path) else in_name

    logger.debug(f"Matching {in_name}")
    # TODO: Should the Beam and field items be relaxed and allowed to be optional?
    # TODOL At very least I think the beam should become options
    # A raw string is used to avoid bad unicode escaping
    regex = re.compile(
        r"^SB(?P<sbid>[0-9]+)"
        r"\.(?P<field>[^.]+)"
        r"((\.beam(?P<beam>[0-9]+))?)"
        r"((\.spw(?P<spw>[0-9]+))?)"
        r"((\.round(?P<round>[0-9]+))?)"
        r"((\.(?P<pol>(i|q|u|v|xx|yy|xy|yx)+))?)"
        r"((\.ch(?P<chl>([0-9]+))-(?P<chh>([0-9]+)))?)"
    )
    results = regex.match(in_name)

    if results is None:
        logger.debug(f"No processed_ms_format results to {in_name} found")
        return None

    groups = results.groupdict()

    logger.debug(f"Matched groups are: {groups}")

    channel_range = (int(groups["chl"]), int(groups["chh"])) if groups["chl"] else None

    return ProcessedNameComponents(
        sbid=groups["sbid"],
        field=groups["field"],
        beam=groups["beam"],
        spw=groups["spw"],
        round=groups["round"],
        pol=groups["pol"],
        channel_range=channel_range,
    )


def create_path_from_processed_name_components(
    processed_name_components: ProcessedNameComponents, parent_path: Path | None = None
) -> Path:
    """Given an input ProcessedNameComponents create the corresponding path

    Args:
        processed_name_components (ProcessedNameComponents): The naming specification to create
        parent_path (Path | None, optional): The parent directory of the output path. Defaults to None.

    Returns:
        Path: A directory with following the specification of the input ProcessedNameComponents
    """

    components = []

    # Operate over each field in order and create the formatted field appropriately
    if processed_name_components.sbid:
        components.append(f"SB{processed_name_components.sbid}")
    if processed_name_components.field:
        components.append(processed_name_components.field)
    if processed_name_components.beam:
        components.append(f"beam{int(processed_name_components.beam):02d}")
    if processed_name_components.spw:
        components.append(f"spw{int(processed_name_components.spw):02d}")
    if processed_name_components.round:
        components.append(f"round{int(processed_name_components.round)}")
    if processed_name_components.pol:
        components.append(f"{processed_name_components.pol}")
    if processed_name_components.channel_range:
        components.append(
            f"ch{processed_name_components.channel_range[0]:04d}-{processed_name_components.channel_range[1]:04d}"
        )

    # Join then add the parent path
    name = ".".join(components)
    out_path = Path(name)

    if parent_path:
        out_path = Path(parent_path) / out_path

    return out_path


def extract_components_from_name(
    name: str | Path,
) -> RawNameComponents | ProcessedNameComponents | CASDANameComponents:
    """Attempts to break down a file name of a recognised format into its principal compobnents.
    Presumably this is a measurement set or something derived from it (i.e. images).

    There are two formats currently recognised:
    - the raw measurement set format produced by the ASKAP ingest system (date, time, beam, spw, underscore delimited)
    - a formatted name produced by flint (SBID, field, beam, spw, dot delimited)

    Internally this function attempts to run two regular expression filters against the input,
    and returns to the set of components that a filter has matched.

    Args:
        name (Union[str,Path]): The name to examine to search for the beam number.

    Raises:
        ValueError: Raised if the name was not was not successfully matched against the known format

    Returns:
        Union[RawNameComponents,ProcessedNameComponents,CASDANamedComponents]: The extracted name components within a name
    """
    name = str(Path(name).name)
    results_raw = raw_ms_format(in_name=name)
    results_formatted = processed_ms_format(in_name=name)
    results_casda = casda_ms_format(in_name=name)

    if all([res is None for res in (results_raw, results_formatted, results_casda)]):
        raise ValueError(f"Unrecognised file name format for {name=}. ")

    matched = [
        res
        for res in (results_raw, results_formatted, results_casda)
        if res is not None
    ]

    if len(matched) > 1:
        logger.info(
            f"The {name=} was recognised as more than one format. Selecting the simplest.  "
        )
        logger.info(f"{results_raw=} {results_formatted=} ")

    results = matched[0]

    return results


def split_images(
    images: list[Path],
    by: str = "pol",
) -> dict[str, list[Path]]:
    """Split a list of images by a given field. This is intended to be used
    when a set of images are to be split by a common field, such as polarisation.

    Args:
        images (List[Path]): The images to split
        by (str, optional): The field to split the images by. Defaults to "pol".

    Returns:
        Dict[str,List[Path]]: A dictionary of the images split by the field
    """
    logger.info(f"Splitting {images=} by {by=}")
    split_dict: dict[str, list[Path]] = {}
    for image in images:
        components = extract_components_from_name(name=image)

        try:
            field = getattr(components, by)
            if field is None:
                raise AttributeError(f"{field=} is None")

        except AttributeError as e:
            msg = f"Failed to extract {by=} from {image=}"
            raise NamingException(msg) from e

        if field not in split_dict:
            split_dict[field] = []
        split_dict[field].append(image)

    return split_dict


def split_and_get_images(
    images: list[Path],
    get: str,
    by: str = "pol",
) -> list[Path]:
    """Split a list of images by a given field and return the images that match
    the field of interest.

    Args:
        images (list[Path]): The images to split
        get (str): The field to extract from the split images
        by (str, optional): How to split the images. Defaults to "pol".

    Raises:
        ValueError: If the field to extract is not found in the split images

    Returns:
        list[Path]: The images that match the field of interest
    """
    split_dict = split_images(images=images, by=by)

    split_list = split_dict.get(get, None)
    if split_list is None:
        raise ValueError(f"Failed to get {get=} from {split_dict=}")

    return split_list


def extract_beam_from_name(name: str | Path) -> int:
    """Attempts to extract the beam number from some input name should it follow a
    known naming convention.

    Args:
        name (Union[str,Path]): The name to examine to search for the beam number.

    Raises:
        ValueError: Raised if the name was not was not successfully matched against the known format

    Returns:
        int: Beam number that extracted from the input name
    """

    results = extract_components_from_name(name=name)
    if results is None or results.beam is None:
        raise ValueError(
            f"Failed to convert to processed name format and find beam: {name=} {results=}"
        )

    return int(results.beam)


def create_ms_name(
    ms_path: Path, sbid: int | None = None, field: str | None = None
) -> str:
    """Create a consistent naming scheme for measurement sets. At present
    it is intended to be used for splitting fields from raw measurement
    sets, but can be expanded.

    Args:
        ms_path (Path): The measurement set being considered. A RawNameComponents will be constructed against it.
        sbid (Optional[int], optional): An explicit SBID to include in the name, otherwise one will attempted to be extracted the the ms path. If these fail the sbid is set of 00000. Defaults to None.
        field (Optional[str], optional): The field that this measurement set will contain. Defaults to None.

    Returns:
        str: The filename of the measurement set
    """

    ms_path = Path(ms_path).absolute()
    ms_name_list: list[Any] = []

    format_components = extract_components_from_name(name=ms_path)

    # Use the explicit SBID is provided, otherwise attempt
    # to extract it
    sbid_text = "SB0000"
    if sbid:
        sbid_text = f"SB{sbid}"
    elif (
        not isinstance(format_components, RawNameComponents) and format_components.sbid
    ):
        sbid_text = f"SB{format_components.sbid}"
    else:
        try:
            sbid = get_sbid_from_path(path=ms_path)
            sbid_text = f"SB{sbid}"
        except Exception as e:
            logger.warning(f"{e}, using default {sbid_text}")
    ms_name_list.append(sbid_text)

    field = (
        field
        if field
        else (
            format_components.field
            if not isinstance(format_components, RawNameComponents)
            and format_components.field
            else None
        )
    )
    if field:
        ms_name_list.append(field)

    if format_components:
        if format_components.beam is not None:
            ms_name_list.append(f"beam{int(format_components.beam):02d}")
        if format_components.spw:
            ms_name_list.append(f"spw{format_components.spw}")

    ms_name_list.append("ms")
    ms_name = ".".join(ms_name_list)

    return ms_name


class AegeanNames(NamedTuple):
    """Base names that would be used in various Aegean related tasks"""

    bkg_image: Path
    """Background map computed by BANE"""
    rms_image: Path
    """RMS noise map computed by BANE"""
    comp_cat: Path
    """Component catalogue produced by the aegean source finder"""
    ds9_region: Path
    """DS9 region overlay file"""
    resid_image: Path
    """Residual map after subtracting component catalogue produced by AeRes"""


def create_aegean_names(base_output: str) -> AegeanNames:
    """Create the expected names for aegean and its tools.

    Args:
        base_output (str): The base name that aegean outputs are built from.

    Returns:
        AegeanNames: A collection of names to be produced by Aegean related tasks
    """
    base_output = str(base_output)

    return AegeanNames(
        bkg_image=Path(f"{base_output}_bkg.fits"),
        rms_image=Path(f"{base_output}_rms.fits"),
        comp_cat=Path(f"{base_output}_comp.fits"),
        ds9_region=Path(f"{base_output}_overlay.reg"),
        resid_image=Path(f"{base_output}_residual.fits"),
    )


class LinmosNames(NamedTuple):
    """Creates expected output names for the yandasoft linmos task."""

    image_fits: Path
    """Path to the final fits co-added image"""
    weight_fits: Path
    """Path to the weights fits image created when co-adding images"""
    parset_output_path: Path
    """Path to the output parset generated"""


def create_linmos_names(
    name_prefix: str | Path, parset_output_path: Path | None = None
) -> LinmosNames:
    """This creates the names that would be output but the yandasoft
    linmos task. It returns the names for the linmos and weight maps
    that linmos would create. These names will have the .fits extension
    with them, but be aware that when the linmos parset if created
    these are omitted.

    Args:
        name_prefix (str | Path): The prefix of the filename that will be used to create the linmos and weight file names.

    Returns:
        LinmosNames: Collection of expected filenames
    """
    name_prefix = str(name_prefix) if isinstance(name_prefix, Path) else name_prefix

    logger.info(f"Linmos name prefix is: {name_prefix}")
    return LinmosNames(
        image_fits=Path(f"{name_prefix}.linmos.fits"),
        weight_fits=Path(f"{name_prefix}.weight.fits"),
        parset_output_path=Path(f"{name_prefix}_parset.txt")
        if parset_output_path is None
        else parset_output_path,
    )


def create_linmos_base_path(
    input_images: list[Path],
    additional_suffixes: str | None = None,
) -> Path:
    """Create the base path of a ``yandasoft linmos`` given a set of input images.
    The default operation is to form the name from the common processed name fields
    amount all of the input images.


    Args:
        input_images (list[Path] | None, optional): If provided the common fields of the input images are used as basis of the path. Defaults to None.
        additional_suffixes (str | None, optional): Any additional suffixes to append. Defaults to None.


    Returns:
        Path: The full path of the parset file
    """

    # Unless something has been specified, we make it up
    logger.info(f"Combining images {input_images}")
    output_name = create_name_from_common_fields(
        in_paths=tuple(input_images), additional_suffixes=additional_suffixes
    )
    out_dir = output_name.parent
    logger.info(f"Base output image name will be: {output_name}")
    assert out_dir is not None, f"{out_dir=}, which should not happen"

    return output_name.absolute()


def get_sbid_from_path(path: Path) -> int:
    """Attempt to extract the SBID of a observation from a path. It is a fairly simple ruleset
    that follows the typical use cases that are actually in practise. There is no mechanism to
    get the SBID from the measurement set meta-data.

    If the path provided ends in a .ms suffix, the parent directory is assumed to be named
    the sbid. Otherwise, the name of the subject directory is. A test is made to ensure the
    sbid is made up of integers only.

    Args:
        path (Path): The path that contains the sbid to extract.

    Raises:
        ValueError: Raised when the SBID extracted from the path is not all digits

    Returns:
        int: The sbid extracted
    """
    path = Path(path)
    path_suffix = path.suffix

    logger.debug(f"Suffix of {path} is {path_suffix}")

    if path_suffix.endswith(".ms"):
        logger.debug("This is a measurement set, so sbid must be the parent directory")
        sbid = path.parent.name
    else:
        sbid = path.name

    if not sbid.isdigit():
        raise ValueError(
            f"Extracted {sbid=} from {path!s} failed appears to be non-conforming - it is not a number! "
        )

    return int(sbid)


def get_potato_output_base_path(ms_path: Path) -> Path:
    """Return the base name for potato peel.

    Args:
        ms_path (Path): Input measurement set that follows the FLINT style process name format

    Returns:
        Path: Output base name to use
    """

    ms_components = processed_ms_format(in_name=ms_path)
    assert ms_components is not None, f"{ms_components=}, which should not be possible"
    output_components = (
        f"SB{ms_components.sbid}.{ms_components.field}.beam{ms_components.beam}.potato"
    )

    output_path = ms_path.parent / output_components
    logger.info(f"Output potato base name: {output_path}")

    return output_path


def get_aocalibrate_output_path(
    ms_path: Path, include_preflagger: bool, include_smoother: bool
) -> Path:
    """Create a name for an aocalibrate style bandpass solution.

    Args:
        ms_path (Path): Path of the measurement set that the aocalibrate file will be created for
        include_preflagger (bool): Whether to include the ``.preflagged`` term to indicate that the preflagger has been executed
        include_smoother (bool): Whether to include the ``.smoothed`` term to included that bandpas smoothing has been performed

    Returns:
        Path: The constructed output path of the solutions file. This include the parent directory of the input measurement set
    """
    ms_components = processed_ms_format(in_name=ms_path)
    assert ms_components is not None, f"{ms_components=}, which should not be possible"
    output_components = [
        f"SB{ms_components.sbid}.{ms_components.field}.beam{ms_components.beam}"
    ]

    if ms_components.spw:
        output_components.append(f"spw{ms_components.spw}")

    output_components.append("aocalibrate")

    if include_preflagger:
        output_components.append("preflagged")
    if include_smoother:
        output_components.append("smoothed")

    output_components.append("bin")

    output_name = ".".join(output_components)
    output_path = ms_path.parent / output_name
    logger.info(f"Constructed {output_path}")

    return output_path


class FITSMaskNames(NamedTuple):
    """Contains the names of the FITS images created when creating a mask image/
    These are only the names, and do not mean that they are necessarily created.
    """

    mask_fits: Path
    """Name of the mask FITS file"""
    signal_fits: Path | None = None
    """Name of the signal FITS file"""


def create_fits_mask_names(
    fits_image: str | Path, include_signal_path: bool = False
) -> FITSMaskNames:
    """Create the names that will be used when generate FITS mask products

    Args:
        fits_image (Union[str,]Path): Base name of the output files
        include_signal_path (bool, optional): If True, also include ``signal_fits`` in the output. Defaults to False.

    Returns:
        FITSMaskNames: collection of names used for the signal and mask FITS images
    """
    fits_image = Path(fits_image)

    fits_signal = (
        fits_image.with_suffix(".signal.fits") if include_signal_path else None
    )
    fits_mask = fits_image.with_suffix(".mask.fits")

    return FITSMaskNames(signal_fits=fits_signal, mask_fits=fits_mask)



from __future__ import annotations

import numpy as np
from scipy.ndimage import median_filter
from scipy.signal import savgol_filter

from flint.logging import logger


def divide_bandpass_by_ref_ant_preserve_phase(
    complex_gains: np.ndarray, ref_ant: int
) -> np.ndarray:
    """Divide the bandpass complex gains (solved for initially by something like
    calibrate) by a nominated reference antenna. In the case of ``calibrate``
    there is no implicit reference antenna. This is valid for cases where the
    xy-phase is set to 0 (true via the ASKAP on-dish calibrator).

    This particular function is most appropriate for the `calibrate` style
    solutions, which solve for the Jones in one step. In HMS notation this
    are normally split into two separate 2x2 matrices, one for the gains
    with zero off-diagonal elements and a leakage matrix with ones on
    the diagonal.

    This is the preferred function to use whena attempting to set a
    phase reference antenna to precomputed Jones bandpass solutions.

    The input complex gains should be in the form:
    >> (ant, channel, pol)

    Internally reference phasores are constructed for the G_x and G_y
    terms of the reference antenna. They are then applied:
    >> G_xp = G_x / G_xref
    >> G_xyp = G_xy / G_yref
    >> G_yxp = G_yx / G_xref
    >> G_y = G_y / G_yref

    which is applied to all antennas in ``complex_gains``.

    Args:
        complex_gains (np.ndarray): The complex gains that will be normalised
        ref_ant (int): The desired reference antenna to use

    Returns:
        np.ndarray: The normalised bandpass solutions
    """
    assert len(complex_gains.shape) == 3, (
        f"The shape of the input complex gains should be of rank 3 in form (ant, chan, pol). Received {complex_gains.shape}"
    )

    logger.info(
        f"Dividing bandpass gain solutions using reference antenna={ref_ant}, using correct phasor"
    )

    # Unpack the values for short hand use
    g_x = complex_gains[:, :, 0]
    g_xy = complex_gains[:, :, 1]
    g_yx = complex_gains[:, :, 2]
    g_y = complex_gains[:, :, 3]

    # In the operations below our ship only wants to be touching
    # the phases in a piratey manner. The amplitudes should remina
    # unchanged. Construct phasors of the nominated reference antenna
    ref_g_x = complex_gains[ref_ant, :, 0]
    ref_g_x = ref_g_x / np.abs(ref_g_x)

    ref_g_y = complex_gains[ref_ant, :, 3]
    ref_g_y = ref_g_y / np.abs(ref_g_y)

    # Now here is the math, from one Captain Daniel Mitchell
    # g_x and g_y.d_yx by g_x(ref) and g_y and g_x.d_xy by g_y(ref).
    # i.e. assuming that xy-phase = 0 (due to the ODC) and that the cross terms are leakage.
    # Since calibrate solves for the Jones directly, the off-diagonals are already
    # multiplied through by the appropriate g_y and g_x.
    g_x_prime = g_x / ref_g_x
    g_xy_prime = g_xy / ref_g_y  # Leakage of y into x, so reference the y
    g_yx_prime = g_yx / ref_g_x  # Leakage of x into y, so reference the x
    g_y_prime = g_y / ref_g_y

    # Construct the output array to slice things into
    bp_p = np.zeros_like(complex_gains) * np.nan

    # Place things into place
    bp_p[:, :, 0] = g_x_prime
    bp_p[:, :, 1] = g_xy_prime
    bp_p[:, :, 2] = g_yx_prime
    bp_p[:, :, 3] = g_y_prime

    return bp_p


def divide_bandpass_by_ref_ant(complex_gains: np.ndarray, ref_ant: int) -> np.ndarray:
    """Divide the bandpass complex gains (solved for initially by something like
    calibrate) by a nominated reference antenna. In the case of ``calibrate``
    there is no implicit reference antenna.

    The input complex gains should be in the form:
    >> (ant, channel, pol)

    Internally this function will construct a phasor:
    >> phasor = Jones_{ref_ant} / abs(Jones_{ref_ant})
    >> shift = Jones_{ref_ant}[0] / and(Jones_{ref_ant}[0])
    >> phasor_shifted = phasor / shift

    which is applied to all antennas in ``complex_gains``. The resulting
    solutions will all have been referenced to the `G_x` of the reference
    antenna. In other words, the phase of all `G_x` items of the reference
    antenna will be zero.

    Args:
        complex_gains (np.ndarray): The complex gains that will be normalised
        ref_ant (int): The desired reference antenna to use

    Returns:
        np.ndarray: The normalised bandpass solutions
    """
    assert len(complex_gains.shape) == 3, (
        f"The shape of the input complex gains should be of rank 3 in form (ant, chan, pol). Received {complex_gains.shape}"
    )

    logger.info(
        f"Dividing bandpass gain solutions using reference antenna={ref_ant} with shifted phasor"
    )

    # Make a copy of the data to avoid editing it wherever else it might be.
    # Trust nothing you pirate.
    complex_gains = complex_gains.copy()

    ref_ant_solutions = complex_gains[ref_ant]

    ref_ant_phasor = ref_ant_solutions / np.abs(ref_ant_solutions)
    ref_ant_shift = (ref_ant_solutions[:, 0] / np.abs(ref_ant_solutions[:, 0]))[:, None]

    # This preserves the consistency of the phase within an single Jones. Essentially
    # setting the reference to the Xx term of the reference antenna.
    phasor_shift = ref_ant_phasor / ref_ant_shift

    logger.info("Multipply, no divide")
    complex_gains = (
        complex_gains / ref_ant_phasor[None, :, :] * phasor_shift[None, :, :]
    )

    return complex_gains


def smooth_data(
    data: np.ndarray,
    window_size: int,
    polynomial_order: int,
    apply_median_filter: bool = False,
) -> np.ndarray:
    """Smooth a 1-dimensional dataset. Internally it uses a savgol filter as
    implemented in scipy.signal.savgol_filter. It is intended to be used to
    smooth the real and imaginary components of the complex gains of the
    bandpass solutions.

    Datapoints that are NaN's are first filled by linearly interpolation the
    closest valid data points. Once the savgol filter has been applied these
    datapoints are then remasked with a NaN.

    If ``median_filter`` is ``True`` then the raw data (without any interpolation)
    will first be passed through a median boxcar filter with a window size of
    ``window_size``.

    Args:
        data (np.ndarray): The 1-dimensional data to be smoothed.
        window_size (int): The size of the window function of the savgol filter. Passed directly to savgol.
        polynomial_order (int): The order of the polynomial of the savgol filter. Passed directly to savgol.
        apply_median_filter (bool, optional): Apply a median filter to the data before applying the savgol filter using the same window size. Defaults to False.

    Returns:
        np.ndarray: Smoothed dataset
    """

    # Make a copy so we do not mess around with the original numpy data
    # where ever it might be. Trust nothing you sea dog.
    data = data.copy()

    if np.all(~np.isfinite(data)):
        return data

    if apply_median_filter:
        data = median_filter(input=data, size=window_size)

    # Before we smooth we need to fill in channels that are flagged with nans.
    # For this we will apply a simply linear interpolation across the blanked
    # regions, smooth, and then reflag them later
    x = np.arange(len(data))
    mask = ~np.isfinite(data)
    d_interp = np.interp(x[mask], x[~mask], data[~mask])

    # This is the filling in of the blanks
    data[mask] = d_interp

    # Now we smooth. This savgol filter fits a polynomial to successive subsets of data
    # in a manner similar to a box car. The final positional argument here denoted the
    # behaviour of the edge where the window (second positional argument) does not have
    # enough data. This process is similar to the original implemented in bptools, except
    # here we are using a polynomial, not a set of harmonic basis functions.
    smoothed_data = savgol_filter(data, window_size, polynomial_order)
    smoothed_data[mask] = np.nan

    return smoothed_data


def smooth_bandpass_complex_gains(
    complex_gains: np.ndarray,
    window_size: int = 16,
    polynomial_order: int = 4,
    apply_median_filter: bool = True,
    smooth_jones_elements: tuple[int, ...] = (0, 1, 2, 3),
) -> np.ndarray:
    """Smooth bandpass solutions by applying a savgol filter to the real and imaginary components
    of each of the antenna based polarisation solutions across channels.

    Experience suggests that this performs better once the data have been divided by a reference
    antenna. Flagged channels (represented by NaNs) will be maintained in the smoothed
    output data.

    The input bandpass data contained by ``complex_gains`` is expected to be in the form:
    > [ants, chans, pols]

    If ``median_filter`` is ``True`` then the raw data (without any interpolation)
    will first be passed through a median boxcar filter with a window size of
    ``window_size``.

    Args:
        complex_gains (np.ndarray): Data to be smoothed.
        window_size (int, optional): The size of the window function of the savgol filter. Passed directly to savgol. Defaults to 16.
        polynomial_order (int, optional): The order of the polynomial of the savgol filter. Passed directly to savgol. Defaults to 4.
        apply_median_filter (bool, optional): Apply a median filter to the data before applying the savgol filter using the same window size. Defaults to True.
        smoothe_jones_elements (Tuple[int, ...], optional): Which elements of the antennae Jones will be smoothed through frequency, i.e. X_x, X_y, Y_x, Y_y. Defaults to (0, 1, 2, 3).

    Returns:
        np.ndarray: Smoothed complex gains
    """

    assert len(complex_gains.shape) == 3, (
        f"The shape of the input complex gains should be of rank 3 in form (ant, chan, pol). Received {complex_gains.shape}"
    )

    # Duplicate the original, ya filthy pirate
    smoothed_complex_gains = complex_gains.copy()

    ants, chans, pols = complex_gains.shape

    logger.info(f"Smoothing using {window_size=} {polynomial_order=}")

    for ant in range(ants):
        # TODO: This will be smoothing the X_y and Y_x. Should this actually be done?
        for pol in smooth_jones_elements:
            assert pol in (0, 1, 2, 3), f"{pol=} is not valid Jones entry. "

            smoothed_complex_gains[ant, :, pol].real = smooth_data(
                data=complex_gains[ant, :, pol].real,
                window_size=window_size,
                polynomial_order=polynomial_order,
                apply_median_filter=apply_median_filter,
            )
            smoothed_complex_gains[ant, :, pol].imag = smooth_data(
                data=complex_gains[ant, :, pol].imag,
                window_size=window_size,
                polynomial_order=polynomial_order,
                apply_median_filter=apply_median_filter,
            )

    return smoothed_complex_gains



"""Tests that work around the catalogue functionality"""

from __future__ import annotations

from pathlib import Path

import pytest
from astropy.table import Table

from flint.catalogue import (
    KNOWN_REFERENCE_CATALOGUES,
    Catalogue,
    _guess_catalogue_type,
    download_referencce_catalogues,
    download_vizier_catalogue,
    get_reference_catalogue,
    guess_column_in_table,
    verify_reference_catalogues,
)
from flint.utils import get_packaged_resource_path


def test_verify_catalogue_check(tmpdir):
    """Makes sure that the verify catalogue function works nicely.
    Only a basic check if it exists is performed, don't try to
    download"""

    catalogue_directory = Path(tmpdir) / "catalogue1"
    catalogue_directory.mkdir(parents=True)

    assert not verify_reference_catalogues(
        reference_directory=catalogue_directory, load_catalogue=False
    )

    for k, v in KNOWN_REFERENCE_CATALOGUES.items():
        file_name = v.file_name
        temp = catalogue_directory / file_name
        temp.touch()

    assert verify_reference_catalogues(
        reference_directory=catalogue_directory, load_catalogue=False
    )


def _get_aegean_catalogue():
    table_path = get_packaged_resource_path(
        package="flint.data.tests",
        filename="SB38959.RACS_1357-18.noselfcal.linmos_comp.fits",
    )
    table = Table.read(table_path)
    return table, table_path


def test_guess_column_names():
    """See if the correct column names can be guessed from a table"""
    table, _ = _get_aegean_catalogue()

    exp = dict(
        ra="ra",
        dec="dec",
        intflux="int_flux",
        peakflux="peak_flux",
        intfluxerr="local_rms",
    )
    for key, value in exp.items():
        assert value == guess_column_in_table(table=table, column=key)
        assert value == guess_column_in_table(
            table=table, column=key, guess_column="JackSparrow"
        )

    with pytest.raises(KeyError):
        guess_column_in_table(table=table, column="JackSparrow")


def test_catalogue_type():
    """Testing the guessing of a catalogue. At the moment this
    guess function is a stub function for the moment."""
    table, table_path = _get_aegean_catalogue()

    cata = _guess_catalogue_type(table=table_path)
    assert isinstance(cata, Catalogue)

    cata2 = _guess_catalogue_type(table=table)
    assert isinstance(cata2, Catalogue)


def test_known_reference_catalogues():
    """Make sure all of the known reference catalogues have a vizier id attached"""
    assert all([cata.vizier_id for cata in KNOWN_REFERENCE_CATALOGUES.values()])


def test_no_reference_catalogue():
    """Ensure file not found error raised if catalogue not found"""
    with pytest.raises(FileNotFoundError):
        _ = get_reference_catalogue(reference_directory=Path("./"), survey="NVSS")


def test_download_vizier_catalogue(tmpdir):
    """Download a example vizier table"""

    output_path = Path(tmpdir) / "catalogue1/ICRF.fits"
    output_path.parent.mkdir(parents=True, exist_ok=True)

    icrf_id = KNOWN_REFERENCE_CATALOGUES["ICRF"]
    assert icrf_id.vizier_id
    cata_path = download_vizier_catalogue(
        output_path=output_path, vizier_id=icrf_id.vizier_id
    )

    assert cata_path == output_path
    assert cata_path.exists()

    table = Table.read(cata_path)
    assert len(table) == 3414


def test_get_vizier_catalogue(tmpdir):
    """Download a example vizier table"""
    output_path = Path(tmpdir) / "catalogue1/ICRF.fits"
    assert not output_path.exists()

    output_path = Path(tmpdir) / "catalogue1/ICRF.fits"
    output_path.parent.mkdir(parents=True, exist_ok=True)

    icrf = KNOWN_REFERENCE_CATALOGUES["ICRF"]
    assert icrf.vizier_id
    _ = download_vizier_catalogue(output_path=output_path, vizier_id=icrf.vizier_id)
    assert output_path.exists()

    table, catalogue = get_reference_catalogue(
        reference_directory=output_path.parent, survey="ICRF"
    )
    assert catalogue.file_name == "ICRF.fits"
    assert catalogue.survey == "ICRF"
    assert len(table) == 3414

    with pytest.raises(ValueError):
        _, _ = get_reference_catalogue(
            reference_directory=output_path.parent, survey="Jack"
        )


def test_download_vizier_catalogue_dryrun(tmpdir):
    """See if the dry run option in download a example vizier table"""

    output_path = Path(tmpdir) / "cataloguedry/ICRF.fits"
    output_path.parent.mkdir(parents=True, exist_ok=True)

    icrf_cata = KNOWN_REFERENCE_CATALOGUES["ICRF"]
    assert icrf_cata.vizier_id is not None

    cata_path = download_vizier_catalogue(
        output_path=output_path, vizier_id=icrf_cata.vizier_id, dry_run=True
    )

    assert cata_path == output_path
    assert not cata_path.exists()


def test_download_reference_catalogues(tmpdir):
    """Ensure all catalogues can be downloaded. Not the dry_run=True,
    meaning the catalogues are not all actually dowenloaded
    """
    output_dir = Path(tmpdir) / "catalogue2"
    output_dir.mkdir(parents=True, exist_ok=True)

    outputs = download_referencce_catalogues(
        reference_directory=output_dir, dry_run=True
    )

    assert len(outputs) == len(KNOWN_REFERENCE_CATALOGUES)
